from bxcommon.feed.raw_block import AbstractRawBlock
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo


class EthRawBlock(AbstractRawBlock[InternalEthBlockInfo]):
    """
    EthBlock Notification Object

    """
    pass



import json
import asyncio
from asyncio import Future
from typing import TYPE_CHECKING, Dict, NamedTuple, Optional, Any, cast, Type, Tuple

from aiohttp import WSMessage
from multidict import CIMultiDictProxy

from bxcommon.feed.feed import FeedKey
from bxcommon.rpc.abstract_rpc_handler import AbstractRpcHandler
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.requests.subscribe_rpc_request import SubscribeRpcRequest
from bxcommon.rpc.requests.unsubscribe_rpc_request import UnsubscribeRpcRequest
from bxcommon.rpc.rpc_errors import RpcParseError
from bxcommon.rpc.rpc_request_type import RpcRequestType

from bxcommon import constants
from bxcommon.feed.feed_manager import FeedManager
from bxcommon.feed.subscriber import Subscriber
from bxutils import log_messages
from bxutils import logging
from bxutils.encoding.json_encoder import Case

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


logger = logging.get_logger(__name__)


class Subscription(NamedTuple):
    subscriber: Subscriber
    feed_key: FeedKey
    task: Future
    account_id: Optional[str] = None


class WsRequest(NamedTuple):
    message: WSMessage
    headers: CIMultiDictProxy


class AbstractWsRpcHandler(AbstractRpcHandler["AbstractNode", WsRequest, str]):
    feed_manager: FeedManager
    subscriptions: Dict[str, Subscription]
    subscribed_messages: "asyncio.Queue[BxJsonRpcRequest]"

    def __init__(self, node: "AbstractNode", feed_manager: FeedManager, case: Case) -> None:
        super().__init__(node, case)
        self.subscriptions = {}

        self.feed_manager = feed_manager
        self.subscribed_messages = asyncio.Queue(
            constants.RPC_SUBSCRIBER_MAX_QUEUE_SIZE
        )
        self.disconnect_event = asyncio.Event()

    async def parse_request(self, request: WsRequest) -> Dict[str, Any]:
        try:
            request_message_dict = {}
            if request.message.data:
                request_message_dict = json.loads(request.message.data)
            assert isinstance(request_message_dict, Dict)
            return request_message_dict
        except Exception:
            raise RpcParseError(None, f"Unable to parse the request: {request}")

    def get_request_handler(self, request: BxJsonRpcRequest) -> AbstractRpcRequest:
        if request.method == RpcRequestType.SUBSCRIBE and request.method in self.request_handlers:
            return self._subscribe_request_factory(request)
        elif request.method == RpcRequestType.UNSUBSCRIBE and request.method in self.request_handlers:
            return self._unsubscribe_request_factory(request)

        request_handler_type = self.request_handlers[request.method]
        # seems to be pyre bug: https://github.com/facebook/pyre-check/issues/267
        # pyre-fixme[45]: Cannot instantiate abstract class `Abstract` with abstract method `run`.
        return request_handler_type(request, self.node)

    def serialize_response(self, response: JsonRpcResponse) -> str:
        return response.to_jsons(self.case)

    def serialize_cached_subscription_message(self, message: BxJsonRpcRequest) -> bytes:
        return self.node.serialized_message_cache.serialize_from_cache(message, self.case)

    async def get_next_subscribed_message(self) -> BxJsonRpcRequest:
        return await self.subscribed_messages.get()

    async def handle_subscription(self, subscriber: Subscriber) -> None:
        while True:
            notification = await subscriber.receive()
            # subscription notifications are sent as JSONRPC requests
            next_message = BxJsonRpcRequest(
                None,
                RpcRequestType.SUBSCRIBE,
                {
                    "subscription": subscriber.subscription_id,
                    "result": notification
                }
            )
            if self.subscribed_messages.full():
                logger.info(
                    log_messages.BAD_RPC_SUBSCRIBER,
                    self.subscribed_messages.qsize(),
                    list(self.subscriptions.keys())
                )
                asyncio.create_task(self.async_close())
                return
            else:
                await self.subscribed_messages.put(next_message)

    async def wait_for_close(self) -> None:
        await self.disconnect_event.wait()

    async def async_close(self) -> None:
        self.close()

    def close(self) -> None:
        subscription_ids = list(self.subscriptions.keys())
        for subscription_id in subscription_ids:
            feed_key, account_id = self._on_unsubscribe(subscription_id)
            assert feed_key is not None
            self.feed_manager.unsubscribe_from_feed(
                feed_key,
                subscription_id,
                account_id
            )
        self.subscriptions = {}

        self.disconnect_event.set()

    def _on_new_subscriber(self, subscriber: Subscriber, feed_key: FeedKey, account_id: Optional[str] = None) -> None:
        task = asyncio.ensure_future(self.handle_subscription(subscriber))
        self.subscriptions[subscriber.subscription_id] = Subscription(
            subscriber, feed_key, task, account_id
        )
        self.node.on_new_subscriber_request()

    def _on_unsubscribe(self, subscriber_id: str) -> Tuple[Optional[FeedKey], Optional[str]]:
        if subscriber_id in self.subscriptions:
            (_, feed_key, task, account_id) = self.subscriptions.pop(subscriber_id)
            task.cancel()
            return feed_key, account_id
        return None, None

    def _subscribe_request_factory(
        self, request: BxJsonRpcRequest
    ) -> AbstractRpcRequest:
        subscribe_rpc_request = cast(Type[SubscribeRpcRequest], self.request_handlers[RpcRequestType.SUBSCRIBE])
        return subscribe_rpc_request(
            request, self.node, self.feed_manager, self._on_new_subscriber
        )

    def _unsubscribe_request_factory(
        self, request: BxJsonRpcRequest
    ) -> AbstractRpcRequest:
        unsubscribe_rpc_request = cast(Type[UnsubscribeRpcRequest], self.request_handlers[RpcRequestType.UNSUBSCRIBE])
        return unsubscribe_rpc_request(
            request, self.node, self.feed_manager, self._on_unsubscribe
        )



import json
from ssl import SSLContext
from typing import Optional, Dict, Any, Union, List

import ast
import status
from urllib3 import Retry, HTTPResponse
from urllib3.exceptions import HTTPError, MaxRetryError
from urllib3.poolmanager import PoolManager
from urllib3.util import parse_url

from bxcommon import constants
from bxutils import log_messages
from bxutils import logging
from bxutils.encoding import json_encoder

# recursive types are not supported: https://github.com/python/typing/issues/182

JT = Union[Dict[str, Any], List[Any]]

logger = logging.get_logger(__name__)

_url = constants.SDN_ROOT_URL
_ssl_context: Optional[SSLContext] = None

METHODS_WHITELIST = frozenset(
    ["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST", "PATCH"]
)


def set_root_url(sdn_url: str, ssl_context: Optional[SSLContext] = None):
    # pylint: disable=global-statement
    global _url
    _url = sdn_url
    update_http_ssl_context(ssl_context)


def update_http_ssl_context(ssl_context: Optional[SSLContext] = None):
    # pylint: disable=global-statement
    global _ssl_context
    _ssl_context = ssl_context


def post_json(endpoint: str, payload=None) -> Optional[JT]:
    return _http_request(
        "POST",
        endpoint,
        False,
        body=json_encoder.to_json(payload),
        headers=constants.HTTP_HEADERS
    )


def patch_json(endpoint: str, payload=None) -> Optional[JT]:
    return _http_request(
        "PATCH",
        endpoint,
        False,
        body=json_encoder.to_json(payload),
        headers=constants.HTTP_HEADERS
    )


def delete_json(endpoint: str, payload=None) -> Optional[JT]:
    return _http_request(
        "DELETE",
        endpoint,
        False,
        body=json_encoder.to_json(payload),
        headers=constants.HTTP_HEADERS
    )


def get_json(endpoint: str) -> Optional[JT]:
    return _http_request(
        "GET",
        endpoint,
        False,
        headers=constants.HTTP_HEADERS
    )


def get_json_raising_timeout(endpoint: str) -> Optional[JT]:
    return _http_request(
        "GET",
        endpoint,
        True,
        headers=constants.HTTP_HEADERS
    )


def get_json_with_payload(endpoint: str, payload=None) -> Optional[JT]:
    return _http_request(
        "GET",
        endpoint,
        False,
        body=json_encoder.to_json(payload),
        headers=constants.HTTP_HEADERS
    )


def build_url(endpoint: str) -> str:
    if not endpoint or not isinstance(endpoint, str):
        raise ValueError("Missing or invalid URL")
    return _url + endpoint


def raise_for_status(res: HTTPResponse) -> None:
    if status.is_client_error(res.status) or status.is_server_error(res.status):
        raise HTTPError(f"{res.status}:{res.reason}. {ast.literal_eval(res.data.decode('utf-8'))}")


def _http_request(
    method: str, endpoint: str, handle_timeout: bool, **kwargs
) -> Optional[JT]:
    url = build_url(endpoint)
    parsed_url = parse_url(url)
    pm_args = {
        "num_pools": constants.HTTP_POOL_MANAGER_COUNT,
        "host": parsed_url.host,
        "port": parsed_url.port,
        "retries": Retry(
            connect=constants.HTTP_REQUEST_RETRIES_COUNT,
            read=constants.HTTP_REQUEST_RETRIES_COUNT,
            redirect=constants.HTTP_REQUEST_RETRIES_COUNT,
            backoff_factor=constants.HTTP_REQUEST_BACKOFF_FACTOR,
            method_whitelist=METHODS_WHITELIST
        ),
        "ssl_context": _ssl_context,
    }
    if _ssl_context is not None and url.startswith("https"):
        pm_args["assert_hostname"] = False
    http_pool_manager: PoolManager = PoolManager(**pm_args)
    try:
        logger.trace("HTTP {0} to {1}", method, url)
        response = http_pool_manager.request(
            method=method,
            url=parsed_url.url,
            timeout=constants.HTTP_REQUEST_TIMEOUT,
            **kwargs
        )
        raise_for_status(response)
    except MaxRetryError as e:
        logger.info("{} to {} failed due to: {}.", method, url, e)
        return None
    except TimeoutError:
        if handle_timeout:
            raise TimeoutError
        return None
    except Exception as e:  # pylint: disable=broad-except
        logger.error(log_messages.HTTP_REQUEST_RETURNED_ERROR, method, url, e)
        return None

    return json.loads(response.data)



import functools
from asyncio import Future
from typing import List, Optional

from mock import MagicMock

from bxcommon import constants
from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.abstract_node import AbstractNode
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.constants import DEFAULT_NETWORK_NUM
from bxcommon.models.node_type import NodeType
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.peer_info import ConnectionPeerInfo
from bxcommon.services.broadcast_service import BroadcastService
from bxcommon.services.transaction_service import TransactionService
from bxcommon.test_utils.mocks.mock_node_ssl_service import MockNodeSSLService
from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.common_opts import CommonOpts
from bxcommon.utils.stats.memory_statistics_service import memory_statistics
from bxutils.services.node_ssl_service import NodeSSLService


class MockNode(AbstractNode):
    NODE_TYPE = NodeType.RELAY

    def __init__(self, opts: CommonOpts, node_ssl_service: Optional[NodeSSLService] = None) -> None:
        if node_ssl_service is None:
            node_ssl_service = MockNodeSSLService(self.NODE_TYPE, MagicMock())
        super(MockNode, self).__init__(opts, node_ssl_service)
        self.alarm_queue = AlarmQueue()
        self.network_num = DEFAULT_NETWORK_NUM

        self.broadcast_messages = []

        self._tx_service = TransactionService(self, self.network_num)
        self._tx_services = {}

    def broadcast(
        self, msg, broadcasting_conn=None, prepend_to_queue=False, connection_types=None
    ) -> List[AbstractConnection]:
        self.broadcast_messages.append(msg)
        return []

    def get_tx_service(self, network_num: Optional[int] = None) -> TransactionService:
        return self._tx_service

    def get_outbound_peer_addresses(self):
        pass

    def get_outbound_peer_info(self) -> List[ConnectionPeerInfo]:
        pass

    def get_broadcast_service(self) -> BroadcastService:
        pass

    def sync_and_send_request_for_relay_peers(self, network_num: int):
        pass

    def build_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> Optional[AbstractConnection]:
        pass

    def process_potential_relays_from_sdn(self, get_potential_relays_future: Future):
        pass

    def on_failed_connection_retry(
        self, ip: str, port: int, connection_type: ConnectionType, connection_state: ConnectionState
    ) -> None:
        pass

    def sync_tx_services(self):
        pass

    def _transaction_sync_timeout(self) -> int:
        pass

    def check_sync_relay_connections(self, conn: AbstractConnection) -> int:
        pass

    def broadcast_transaction(
        self,
        message,
        broadcasting_connection,
        tx_status,
        prepend_to_queue: bool = False
    ) -> None:
        pass

    def on_new_subscriber_request(self) -> None:
        pass

    def init_memory_stats_logging(self):
        memory_statistics.set_node(self)
        memory_statistics.start_recording(
            functools.partial(
                self.record_mem_stats,
                constants.GC_LOW_MEMORY_THRESHOLD,
                constants.GC_MEDIUM_MEMORY_THRESHOLD,
                constants.GC_HIGH_MEMORY_THRESHOLD
            )
        )

    def reevaluate_transaction_streamer_connection(self) -> None:
        pass



class ThroughputEvent:
    def __init__(self, direction, msg_type, msg_size, peer_desc) -> None:
        self.direction = direction
        self.msg_type = msg_type
        self.msg_size = msg_size
        self.peer_desc = peer_desc



from typing import Union, Tuple, List

from mock import MagicMock

from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage
from bxgateway.messages.ont.tx_ont_message import TxOntMessage
from bxgateway.services.gateway_transaction_service import GatewayTransactionService
from bxgateway.testing.abstract_gateway_transaction_service_test import TestAbstractGatewayTransactionService
from bxgateway.testing.mocks import mock_eth_messages


class TestAbstractEthGatewayTransactionService(TestAbstractGatewayTransactionService):

    def _get_transaction_service(self) -> GatewayTransactionService:
        return GatewayTransactionService(self.node, 0)

    def _get_gateway_node(self) -> AbstractGatewayNode:
        mock_ssl_service = MagicMock()
        return EthGatewayNode(self.opts, mock_ssl_service)

    def _get_node_tx_message(
        self
    ) -> Tuple[
        Union[TxOntMessage, TransactionsEthProtocolMessage], List[Tuple[Sha256Hash, Union[bytearray, memoryview]]]
    ]:
        txs = [
            mock_eth_messages.get_dummy_transaction(1),
            mock_eth_messages.get_dummy_access_list_transaction(2)
        ]

        msg = TransactionsEthProtocolMessage(None, txs)

        # pyre-fixme [7]: Expected `Tuple[Union[TransactionsEthProtocolMessage, TxOntMessage], List[Tuple[Sha256Hash, Union[bytearray, memoryview]]]]`
        return msg, list(map(lambda tx: (tx.hash(), tx.contents()), msg.get_transactions()))



import sys

from bxcommon.common_opts import CommonOpts
from bxcommon.services import sdn_http_service
from bxcommon.utils import cli, model_loader, node_cache
from bxcommon.models.node_model import NodeModel

from bxutils.ssl import ssl_serializer
from bxutils.ssl.ssl_certificate_type import SSLCertificateType
from bxutils.services.node_ssl_service import NodeSSLService
from bxutils import logging


logger = logging.get_logger(__name__)


def _register_node(opts: CommonOpts, node_ssl_service: NodeSSLService) -> NodeModel:
    temp_node_model = model_loader.load_model(NodeModel, opts.__dict__)
    if node_ssl_service.should_renew_node_certificate():
        temp_node_model.csr = ssl_serializer.serialize_csr(
            node_ssl_service.create_csr()
        )

    try:
        node_model = sdn_http_service.register_node(temp_node_model)
    except ValueError as e:
        logger.fatal(e)
        sys.exit(1)
    except EnvironmentError as e:
        logger.info(
            "Unable to contact SDN to register node using {}, attempting to get information from cache",
            opts.sdn_url,
        )

        cache_info = node_cache.read(opts)
        if not cache_info or not cache_info.node_model:
            logger.fatal(
                "Unable to reach the SDN and no local cache information was found. Unable to start the node"
            )
            sys.exit(1)
        node_model = cache_info.node_model

    if node_model.should_update_source_version:
        logger.info(
            "UPDATE AVAILABLE! An updated software version is available, please download and install the "
            "latest version"
        )

    return node_model


def set_network_info(opts: CommonOpts, _node_ssl_service) -> None:
    cli.set_blockchain_networks_info(opts)
    cli.parse_blockchain_opts(opts, opts.node_type)


def set_node_model(opts: CommonOpts, node_ssl_service) -> None:
    """
    Must be executed after set_network_info
    :param self:
    :param opts:
    :param node_ssl_service:
    :return:
    """
    node_model = None
    if opts.node_id:
        # Test network, get pre-configured peers from the SDN.
        node_model = sdn_http_service.fetch_node_attributes(opts.node_id)

    if not node_model:
        node_model = _register_node(opts, node_ssl_service)

    if node_model.cert is not None:
        private_cert = ssl_serializer.deserialize_cert(node_model.cert)
        node_ssl_service.blocking_store_node_certificate(private_cert)
        ssl_context = node_ssl_service.create_ssl_context(SSLCertificateType.PRIVATE)
        sdn_http_service.reset_pool(ssl_context)

    # Add opts from SDN, but don't overwrite CLI args
    default_values_to_update = [None, -1]
    for key, val in node_model.__dict__.items():
        if opts.__dict__.get(key) in default_values_to_update:
            opts.__dict__[key] = val


init_tasks = [
    set_network_info,
    set_node_model,
]



from typing import Optional, Dict, Any, List, Collection

import blxr_rlp as rlp

from bxcommon.messages.eth.serializers.transaction_type import EthTransactionType
from bxcommon.messages.eth.serializers.unsigned_transaction import UnsignedTransaction
from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.eth import eth_common_utils, eth_common_constants, crypto_utils
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import utils


class AccessedAddress(rlp.Serializable):
    fields = [
        (
            "address",
            rlp.sedes.Binary.fixed_length(eth_common_constants.ADDRESS_LEN, allow_empty=True),
        ),
        ("storage_keys", rlp.sedes.CountableList(rlp.sedes.binary)),
    ]

    address: bytearray
    storage_keys: List[bytearray]

    def __init__(self, *args, **kwargs):
        self.address = bytearray()
        self.storage_keys = []

        super().__init__(*args, **kwargs)

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "AccessedAddress":
        return AccessedAddress(
            utils.or_else(
                utils.optional_map(payload["address"], lambda addr: convert.hex_to_bytes(addr[2:])),
                bytes(),
            ),
            [
                utils.optional_map(key, lambda k: convert.hex_to_bytes(k[2:]))
                for key in payload["storageKeys"]
            ],
        )


class Transaction(rlp.Serializable):
    """
    Some notes on the Berlin implementation details:
    (see https://eips.ethereum.org/EIPS/eip-2718 for spec)

    Transaction can either be an RLP encoded list of all transaction attributes
    (e.g. `[nonce, gas_price, start_gas, ..., v, r, s]`), or an opaque byte string
    consisting of `"[transaction_type][nonce, gas_price, start_gas, ...]"`.

    As a result, `Transaction.serialize/deserialize` can return either the
    RLP list or the opaque byte string, and this can be confusing to work with.
    As a rule of thumb, LegacyTransactions will need to call `rlp.encode` to
    get the transaction byte representation for purposes such as calculating its
    hash, while new transaction should not, as that will result in double encoding.

    It's possible for LegacyTransaction to be in either representation. In our
    implementation, we currently choose to always serialize LegacyTransaction in
    the original, first representations, since this is easier on test cases to
    ensure we have a mix of old and new transactions in blocks. However, this
    code is capable of understanding LegacyTransaction in either format.
    """

    transaction_type: EthTransactionType = EthTransactionType.LEGACY

    nonce: int = 0
    gas_price: int = 0
    start_gas: int = 0
    to: Optional[bytearray] = None
    value: int = 0
    data: bytearray
    v: int = 0  # pylint: disable=invalid-name
    r: int = 0  # pylint: disable=invalid-name
    s: int = 0  # pylint: disable=invalid-name

    def __init__(self, *args, **kwargs):
        self.data = bytearray()
        super().__init__(*args, **kwargs)

    # pylint: disable=arguments-differ
    @classmethod
    def serialize(cls, obj, type_parsed: bool = False, **kwargs):
        if type_parsed:
            result = super().serialize(obj)
            if obj.transaction_type == EthTransactionType.LEGACY:
                return result
            return obj.transaction_type.encode_rlp() + rlp.encode(result)
        else:
            return obj.__class__.serialize(obj, type_parsed=True, **kwargs)

    # pylint: disable=arguments-differ
    @classmethod
    def deserialize(cls, serial, type_parsed: bool = False, **extra_kwargs):
        if type_parsed:
            return super().deserialize(serial, **extra_kwargs)

        if isinstance(serial, (list, tuple)):
            return LegacyTransaction.deserialize(serial, type_parsed=True, **extra_kwargs)

        if isinstance(serial, memoryview):
            serial = serial.tobytes()

        if isinstance(serial, bytes):
            transaction_flag = serial[0]
            if transaction_flag <= eth_common_constants.MAX_TRANSACTION_TYPE:
                transaction_type = EthTransactionType(serial[0])
                payload = rlp.decode(serial[1:])
            else:
                payload = rlp.decode(serial)
                transaction_type = EthTransactionType.LEGACY

            if transaction_type == EthTransactionType.LEGACY:
                return LegacyTransaction.deserialize(payload, type_parsed=True, **extra_kwargs)
            elif transaction_type == EthTransactionType.ACCESS_LIST:
                return AccessListTransaction.deserialize(payload, type_parsed=True, **extra_kwargs)
            elif transaction_type == EthTransactionType.DYNAMIC_FEE:
                return DynamicFeeTransaction.deserialize(payload, type_parsed=True, **extra_kwargs)
            else:
                raise ValueError(f"Unexpected transaction type: {transaction_type}")
        raise ValueError(f"Unexpected serial type: {type(serial)}")

    def hash(self) -> Sha256Hash:
        pass

    def contents(self) -> memoryview:
        return memoryview(rlp.encode(self))

    def is_eip_155_signed(self) -> bool:
        return self.v >= eth_common_constants.EIP155_CHAIN_ID_OFFSET

    def chain_id(self) -> int:
        if self.v % 2 == 0:
            v = self.v - 1  # pylint: disable=invalid-name
        else:
            v = self.v  # pylint: disable=invalid-name
        return (v - eth_common_constants.EIP155_CHAIN_ID_OFFSET) // 2

    def signature(self) -> bytes:
        return crypto_utils.encode_signature(self.v, self.r, self.s)

    def get_unsigned(self) -> bytes:
        pass

    def adjusted_gas_price(self, _base_fee: Optional[int] = None) -> int:
        return self.gas_price

    def gas_prices(self) -> Collection[int]:
        """
        Returns a compatible struct for checking all fees have increased for transaction replacement.
        """
        return self.gas_price, self.gas_price

    def to_json(self) -> Dict[str, Any]:
        """
        Serializes data to be close to Ethereum RPC spec for publishing to the transaction
        feed.

        see https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gettransactionbyhash

        Some fields are excluded, since they will never be populated by bxgateway.
        (mainly fields related to the block the transaction gets included in)
        - blockHash
        - blockNumber
        - transactionIndex
        """
        message_hash = self.hash()

        input_data = convert.bytes_to_hex(self.data)
        if not input_data:
            input_data = "0x"
        else:
            input_data = f"0x{input_data}"

        serialized_output = {
            "from": self.from_address(),
            "gas": hex(self.start_gas),
            "gas_price": hex(self.gas_price),
            "hash": f"0x{str(message_hash)}",
            "input": input_data,
            "nonce": hex(self.nonce),
            "value": hex(self.value),
            "v": hex(self.v),
            "r": hex(self.r),
            "s": hex(self.s),
            "type": f"0x{self.transaction_type.value}",
        }

        to = self.to
        if to is not None:
            serialized_output["to"] = convert.bytes_to_hex_string_format(to)

        return serialized_output

    def from_address(self) -> str:
        from_key = crypto_utils.recover_public_key(
            self.get_unsigned(), self.signature(), eth_common_utils.keccak_hash
        )
        from_address = crypto_utils.public_key_to_address(from_key)
        return convert.bytes_to_hex_string_format(from_address)

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "Transaction":

        transaction_cls = LegacyTransaction

        try:
            transaction_type = EthTransactionType(int(payload.get("type", "0x0"), 16))
            if transaction_type == EthTransactionType.ACCESS_LIST:
                transaction_cls = AccessListTransaction
            elif transaction_type == EthTransactionType.DYNAMIC_FEE:
                transaction_cls = DynamicFeeTransaction
        except ValueError:
            # assume legacy transaction if transaction_type access fails
            pass

        return transaction_cls.from_json(payload)

    @classmethod
    def from_json_with_validation(cls, payload: Dict[str, Any]) -> "Transaction":
        """
        create a Transaction from a payload dict.
        this method support a less strict payload. and support input in both the Eth format and our own.
        """
        if "gas_price" in payload:
            payload["gasPrice"] = payload["gas_price"]
        if "access_list" in payload:
            payload["accessList"] = payload["access_list"]
        if "chain_id" in payload:
            payload["chainId"] = payload["chain_id"]
        if "max_priority_fee_per_gas" in payload:
            payload["maxPriorityFeePerGas"] = payload["max_priority_fee_per_gas"]
        if "max_fee_per_gas" in payload:
            payload["maxFeePerGas"] = payload["max_fee_per_gas"]

        for item in ["nonce", "gasPrice", "gas", "value", "v", "r", "s"]:
            value = payload[item]
            if isinstance(value, int):
                payload[item] = hex(value)

        for item in ["accessList", "chainId"]:
            if item in payload:
                value = payload[item]
                if isinstance(value, int):
                    payload[item] = hex(value)

        return cls.from_json(payload)


class LegacyTransaction(Transaction):
    transaction_type: EthTransactionType = EthTransactionType.LEGACY

    fields = [
        ("nonce", rlp.sedes.big_endian_int),
        ("gas_price", rlp.sedes.big_endian_int),
        ("start_gas", rlp.sedes.big_endian_int),
        ("to", rlp.sedes.Binary.fixed_length(eth_common_constants.ADDRESS_LEN, allow_empty=True)),
        ("value", rlp.sedes.big_endian_int),
        ("data", rlp.sedes.binary),
        ("v", rlp.sedes.big_endian_int),
        ("r", rlp.sedes.big_endian_int),
        ("s", rlp.sedes.big_endian_int),
    ]

    def hash(self):
        hash_bytes = eth_common_utils.keccak_hash(rlp.encode(self))
        return Sha256Hash(hash_bytes)

    def get_unsigned(self) -> bytes:
        """
        Returns unsigned transaction.

        EIP-155 protected transactions require the chain ID encoded in the v
        field, and the r/s fields to be empty.
        :return:
        """
        if self.is_eip_155_signed():
            parts = rlp.decode(rlp.encode(Transaction.serialize(self)))
            parts_for_signing = parts[:-3] + [
                eth_common_utils.int_to_big_endian(self.chain_id()),
                b"",
                b"",
            ]
            return rlp.encode(parts_for_signing)
        else:
            return rlp.encode(
                UnsignedTransaction(
                    self.nonce, self.gas_price, self.start_gas, self.to, self.value, self.data
                )
            )

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "Transaction":
        return LegacyTransaction(
            int(payload["nonce"], 16),
            int(payload["gasPrice"], 16),
            int(payload["gas"], 16),
            utils.or_else(
                utils.optional_map(payload["to"], lambda to: convert.hex_to_bytes(to[2:])), bytes()
            ),
            int(payload["value"], 16),
            convert.hex_to_bytes(payload["input"][2:]),
            int(payload["v"], 16),
            int(payload["r"], 16),
            int(payload["s"], 16),
        )


class AccessListTransaction(Transaction):
    transaction_type: EthTransactionType = EthTransactionType.ACCESS_LIST

    fields = [
        ("_chain_id", rlp.sedes.big_endian_int),
        ("nonce", rlp.sedes.big_endian_int),
        ("gas_price", rlp.sedes.big_endian_int),
        ("start_gas", rlp.sedes.big_endian_int),
        ("to", rlp.sedes.Binary.fixed_length(eth_common_constants.ADDRESS_LEN, allow_empty=True)),
        ("value", rlp.sedes.big_endian_int),
        ("data", rlp.sedes.binary),
        ("access_list", rlp.sedes.CountableList(AccessedAddress)),
        ("v", rlp.sedes.big_endian_int),
        ("r", rlp.sedes.big_endian_int),
        ("s", rlp.sedes.big_endian_int),
    ]

    _chain_id: int = 0
    access_list: List[AccessedAddress]

    def __init__(self, *args, **kwargs):
        self.access_list = []
        super().__init__(*args, **kwargs)

    def hash(self):
        hash_bytes = eth_common_utils.keccak_hash(Transaction.serialize(self))
        return Sha256Hash(hash_bytes)

    def chain_id(self) -> int:
        return self._chain_id

    def get_unsigned(self) -> bytes:
        """
        Returns unsigned transaction. EIP-2930 transaction are always EIP-155
        protected. They do not require any of the v/r/s values included.
        :return:
        """

        parts = rlp.decode(Transaction.serialize(self)[1:])
        return self.transaction_type.encode_rlp() + rlp.encode(parts[:-3])

    def signature(self) -> bytes:
        return crypto_utils.encode_signature_y_parity(self.v, self.r, self.s)

    def to_json(self) -> Dict[str, Any]:
        serialized_transaction = super().to_json()
        serialized_transaction["chain_id"] = f"0x{self.chain_id()}"

        access_list = self.access_list
        accessed_addresses = []
        for accessed_address in access_list:
            accessed_addresses.append(
                {
                    "address": convert.bytes_to_hex_string_format(accessed_address.address),
                    "storage_keys": [
                        convert.bytes_to_hex_string_format(key)
                        for key in accessed_address.storage_keys
                    ],
                }
            )

        serialized_transaction.update(
            {
                "access_list": accessed_addresses,
            }
        )

        return serialized_transaction

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "Transaction":
        return AccessListTransaction(
            int(payload["chainId"], 16),
            int(payload["nonce"], 16),
            int(payload["gasPrice"], 16),
            int(payload["gas"], 16),
            utils.or_else(
                utils.optional_map(payload["to"], lambda to: convert.hex_to_bytes(to[2:])), bytes()
            ),
            int(payload["value"], 16),
            convert.hex_to_bytes(payload["input"][2:]),
            [
                AccessedAddress.from_json(accessed_address)
                for accessed_address in payload["accessList"]
            ],
            int(payload["v"], 16),
            int(payload["r"], 16),
            int(payload["s"], 16),
        )


class DynamicFeeTransaction(AccessListTransaction):
    transaction_type: EthTransactionType = EthTransactionType.DYNAMIC_FEE

    fields = [
        ("_chain_id", rlp.sedes.big_endian_int),
        ("nonce", rlp.sedes.big_endian_int),
        ("max_priority_fee_per_gas", rlp.sedes.big_endian_int),     # GasTipCap
        ("max_fee_per_gas", rlp.sedes.big_endian_int),  # GasFeeCap
        ("start_gas", rlp.sedes.big_endian_int),
        ("to", rlp.sedes.Binary.fixed_length(eth_common_constants.ADDRESS_LEN, allow_empty=True)),
        ("value", rlp.sedes.big_endian_int),
        ("data", rlp.sedes.binary),
        ("access_list", rlp.sedes.CountableList(AccessedAddress)),
        ("v", rlp.sedes.big_endian_int),
        ("r", rlp.sedes.big_endian_int),
        ("s", rlp.sedes.big_endian_int),
    ]

    max_priority_fee_per_gas: int = 0
    max_fee_per_gas: int = 0

    def adjusted_gas_price(self, base_fee: Optional[int] = None) -> int:
        """
        Actual gas price, with base fee taken into account. In cases where the base fee is not
        available, only the max_fee_per_gas will be used, though this should not be treated as
        completely accurate.

        see https://ethereum.org/en/developers/docs/gas/#gas-price-london-update
        """
        if base_fee is None:
            return self.max_fee_per_gas

        remainder = self.max_fee_per_gas - base_fee
        tip = min(remainder, self.max_priority_fee_per_gas)
        return base_fee + tip

    def gas_prices(self) -> Collection[int]:
        return self.max_fee_per_gas, self.max_priority_fee_per_gas

    def to_json(self) -> Dict[str, Any]:
        serialized_transaction = super().to_json()
        serialized_transaction.update(
            {
                "gas_price": None,
                "max_priority_fee_per_gas": hex(self.max_priority_fee_per_gas),
                "max_fee_per_gas": hex(self.max_fee_per_gas),
            }
        )

        return serialized_transaction

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "Transaction":
        return DynamicFeeTransaction(
            int(payload["chainId"], 16),
            int(payload["nonce"], 16),
            int(payload["maxPriorityFeePerGas"], 16),
            int(payload["maxFeePerGas"], 16),
            int(payload["gas"], 16),
            utils.or_else(
                utils.optional_map(payload["to"], lambda to: convert.hex_to_bytes(to[2:])), bytes()
            ),
            int(payload["value"], 16),
            convert.hex_to_bytes(payload["input"][2:]),
            [
                AccessedAddress.from_json(accessed_address)
                for accessed_address in payload["accessList"]
            ],
            int(payload["v"], 16),
            int(payload["r"], 16),
            int(payload["s"], 16),
        )


# pylint: disable=protected-access
# RLP hack to allow proper inheritance in sequential calls
DynamicFeeTransaction._sedes = rlp.sedes.List(sedes for _, sedes in DynamicFeeTransaction.fields)



import socket
import struct

from bxutils import logging

from bxcommon import constants

logger = logging.get_logger(__name__)

PORT_PACKING_FORMAT = "<H"


def pack_ip_port(buf, offset, ip, port):
    try:
        buf[offset + 10] = 0xff
        buf[offset + 11] = 0xff
        buf[offset + constants.IP_V4_PREFIX_LENGTH:offset + constants.IP_ADDR_SIZE_IN_BYTES] = \
            socket.inet_pton(socket.AF_INET, ip)
        struct.pack_into(PORT_PACKING_FORMAT, buf, offset + constants.IP_ADDR_SIZE_IN_BYTES, port)
    except socket.error:
        try:
            buf[offset: offset + 16] = socket.inet_pton(socket.AF_INET6, ip)
            struct.pack_into(PORT_PACKING_FORMAT, buf, offset + 16, port)
        except socket.error as e:
            logger.debug("Packing ip port failed!: {}", e)


def unpack_ip_port(buf):
    port, = struct.unpack_from(PORT_PACKING_FORMAT, buf, constants.IP_ADDR_SIZE_IN_BYTES)
    if buf[:+ constants.IP_V4_PREFIX_LENGTH] == constants.IP_V4_PREFIX[:]:
        ip_bytes = buf[constants.IP_V4_PREFIX_LENGTH:constants.IP_ADDR_SIZE_IN_BYTES]
        return socket.inet_ntop(socket.AF_INET, ip_bytes), port
    else:
        return socket.inet_ntop(socket.AF_INET6, buf[:constants.IP_ADDR_SIZE_IN_BYTES]), port



from typing import NamedTuple

from bxcommon import constants
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.utils.expiring_dict import ExpiringDict
from bxutils.encoding.json_encoder import Case


class SerializedMessageKey(NamedTuple):
    case: Case
    object_id: int


class SerializedMessageCache:
    _cache: ExpiringDict[SerializedMessageKey, str]

    def __init__(self, alarm_queue: AlarmQueue) -> None:
        self._cache = ExpiringDict(
            alarm_queue,
            constants.SERIALIZED_MESSAGE_CACHE_EXPIRE_TIME_S,
            "serialized_message_cache"
        )

    def serialize_from_cache(self, message: BxJsonRpcRequest, case: Case) -> bytes:
        if constants.USE_SERIALIZED_MESSAGE_CACHE:
            # this functionality is locked behind a feature flag until changes get
            # submitted to orjson for partial serialization functionality
            # otherwise, cached result gets double-serialized as a string, which
            # is undesirable
            params = message.params
            assert isinstance(params, dict)
            key = SerializedMessageKey(case, id(params["result"]))
            if key in self._cache:
                cached_result = self._cache[key]
                result = message.to_json_bytes_with_cached_result(case, cached_result)
                return result
            else:
                message_bytes, cached_result = message.to_json_bytes_split_serialization(case)
                self._cache.add(key, cached_result)
                return message_bytes
        else:
            return message.to_json_bytes(case)



from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import NULL_ONT_BLOCK_HASH, OntObjectHash
from bxgateway import ont_constants
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.connections.ont.ont_node_connection import OntNodeConnection
from bxgateway.connections.ont.ont_relay_connection import OntRelayConnection
from bxgateway.connections.ont.ont_remote_connection import OntRemoteConnection
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.abstract_block_queuing_service import AbstractBlockQueuingService
from bxgateway.services.ont.ont_block_processing_service import OntBlockProcessingService
from bxgateway.services.ont.ont_block_queuing_service import OntBlockQueuingService
from bxgateway.services.ont.ont_neutrality_service import OntNeutralityService
from bxgateway.services.ont.ont_normal_block_cleanup_service import OntNormalBlockCleanupService
from bxutils.services.node_ssl_service import NodeSSLService

import bxgateway.messages.ont.ont_message_converter_factory as block_msg_converter_factory
import bxgateway.messages.ont.ont_consensus_message_converter_factory as consensus_msg_converter_factory


class OntGatewayNode(AbstractGatewayNode):
    def __init__(self, opts, node_ssl_service: NodeSSLService):
        super(OntGatewayNode, self).__init__(opts, node_ssl_service, ont_constants.TRACKED_BLOCK_CLEANUP_INTERVAL_S)

        self.block_processing_service = OntBlockProcessingService(self)

        self.message_converter = block_msg_converter_factory.create_ont_message_converter(
            self.opts.blockchain_net_magic,
            self.opts
        )
        self.consensus_message_converter = consensus_msg_converter_factory.create_ont_consensus_message_converter(
            self.opts.blockchain_net_magic, self.opts
        )

        self.neutrality_service = OntNeutralityService(self)

        self.current_block_height = 0
        self.current_block_hash = NULL_ONT_BLOCK_HASH
        self.opts.has_fully_updated_tx_service = True

    def build_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return OntNodeConnection(socket_connection, self)

    def build_relay_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> AbstractRelayConnection:
        return OntRelayConnection(socket_connection, self)

    def build_remote_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return OntRemoteConnection(socket_connection, self)

    def build_block_queuing_service(
        self,
        connection: AbstractGatewayBlockchainConnection
    ) -> AbstractBlockQueuingService:
        return OntBlockQueuingService(self, connection)

    def build_block_cleanup_service(self) -> AbstractBlockCleanupService:
        if self.opts.use_extensions:
            from bxgateway.services.ont.ont_extension_block_cleanup_service import OntExtensionBlockCleanupService
            block_cleanup_service = OntExtensionBlockCleanupService(self, self.network_num)
        else:
            block_cleanup_service = OntNormalBlockCleanupService(self, self.network_num)
        return block_cleanup_service

    def update_current_block_height(self, new_block_height: int, new_block_hash: OntObjectHash):
        if new_block_height >= self.current_block_height:
            self.current_block_height = new_block_height
            self.current_block_hash = new_block_hash




import time
import datetime
from typing import Tuple

import task_pool_executor as tpe


from bxcommon.utils import convert, crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.proxy import task_pool_proxy
from bxcommon.utils.proxy.task_queue_proxy import TaskQueueProxy
from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.utils.errors import message_conversion_error
from bxgateway.utils.block_info import BlockInfo
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxgateway.abstract_message_converter import BlockDecompressionResult
from bxgateway.messages.eth.eth_abstract_message_converter import EthAbstractMessageConverter
from bxutils import logging

logger = logging.get_logger(__name__)


class EthExtensionMessageConverter(EthAbstractMessageConverter):

    DEFAULT_BLOCK_SIZE = eth_common_constants.ETH_DEFAULT_BLOCK_SIZE
    MINIMAL_SUB_TASK_TX_COUNT = eth_common_constants.ETH_MINIMAL_SUB_TASK_TX_COUNT

    def __init__(self):
        super().__init__()
        self._default_block_size = self.DEFAULT_BLOCK_SIZE
        self.compression_tasks = TaskQueueProxy(self._create_compression_task)
        self.decompression_tasks = TaskQueueProxy(self._create_decompression_task)

    def block_to_bx_block(
        self,
        block_msg: InternalEthBlockInfo,
        tx_service: ExtensionTransactionService,
        enable_block_compression: bool,
        min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        compress_start_datetime = datetime.datetime.utcnow()
        compress_start_timestamp = time.time()
        self._default_block_size = max(self._default_block_size, len(block_msg.rawbytes()))
        tsk = self.compression_tasks.borrow_task()
        tsk.init(tpe.InputBytes(block_msg.rawbytes()), tx_service.proxy, enable_block_compression, min_tx_age_seconds)
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            self.compression_tasks.return_task(tsk)
            raise message_conversion_error.eth_block_compression_error(block_msg.block_hash(), e)
        bx_block = tsk.bx_block()
        starting_offset = tsk.starting_offset()
        block = memoryview(bx_block)[starting_offset:]
        compressed_size = len(block)
        original_size = len(block_msg.rawbytes()) - starting_offset
        block_hash = block_msg.block_hash()

        block_info = BlockInfo(
            block_hash,
            tsk.short_ids(),
            compress_start_datetime,
            datetime.datetime.utcnow(),
            (time.time() - compress_start_timestamp) * 1000,
            tsk.txn_count(),
            tsk.compressed_block_hash().hex_string(),
            tsk.prev_block_hash().hex_string(),
            original_size,
            compressed_size,
            100 - float(compressed_size) / original_size * 100,
            tsk.ignored_short_ids()
        )
        self.compression_tasks.return_task(tsk)
        return block, block_info

    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        decompress_start_datetime = datetime.datetime.utcnow()
        decompress_start_timestamp = time.time()
        tsk = self.decompression_tasks.borrow_task()
        tsk.init(tpe.InputBytes(bx_block_msg), tx_service.proxy)
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            block_hash = Sha256Hash(convert.hex_to_bytes(tsk.block_hash().hex_string()))
            self.decompression_tasks.return_task(tsk)
            # TODO find a better solution
            raise message_conversion_error.eth_block_decompression_error(block_hash, e)

        total_tx_count = tsk.tx_count()
        unknown_tx_hashes = [Sha256Hash(bytearray(unknown_tx_hash.binary()))
                             for unknown_tx_hash in tsk.unknown_tx_hashes()]
        unknown_tx_sids = tsk.unknown_tx_sids()
        block_hash = Sha256Hash(convert.hex_to_bytes(tsk.block_hash().hex_string()))

        if tsk.success():
            starting_offset = tsk.starting_offset()
            block = memoryview(tsk.block_message())[starting_offset:]
            block_msg = InternalEthBlockInfo(block)
            content_size = len(block_msg.rawbytes())
            logger.debug("Successfully parsed block broadcast message. {} "
                         "transactions in block {}", total_tx_count, block_hash)
            bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(bx_block_msg))
            compressed_size = len(bx_block_msg)

            block_info = BlockInfo(
                block_hash,
                tsk.short_ids(),
                decompress_start_datetime,
                datetime.datetime.utcnow(),
                (time.time() - decompress_start_timestamp) * 1000,
                total_tx_count,
                bx_block_hash,
                convert.bytes_to_hex(block_msg.prev_block_hash().binary),
                len(block_msg.rawbytes()),
                compressed_size,
                100 - float(compressed_size) / content_size * 100,
                []
            )
        else:
            block_msg = None

            logger.debug(
                "Block recovery needed for {}. Missing {} sids, {} tx hashes. "
                "Total txs in block: {}",
                block_hash,
                len(unknown_tx_sids),
                len(unknown_tx_hashes),
                total_tx_count
            )
            block_info = BlockInfo(
                block_hash,
                tsk.short_ids(),
                decompress_start_datetime,
                datetime.datetime.utcnow(),
                (time.time() - decompress_start_timestamp) * 1000,
                None,
                None,
                None,
                None,
                None,
                None,
                []
            )

        self.decompression_tasks.return_task(tsk)
        return BlockDecompressionResult(block_msg, block_info, unknown_tx_sids, unknown_tx_hashes)

    def _create_compression_task(self) -> tpe.EthBlockCompressionTask:
        return tpe.EthBlockCompressionTask(self._default_block_size, self.MINIMAL_SUB_TASK_TX_COUNT)

    def _create_decompression_task(self) -> tpe.EthBlockDecompressionTask:
        return tpe.EthBlockDecompressionTask(self._default_block_size, self.MINIMAL_SUB_TASK_TX_COUNT)



import argparse
import datetime
import json
import requests
import sys
import time
from collections import defaultdict
from enum import Enum
from typing import Optional

from eth_account import Account
from eth_account.datastructures import SignedTransaction
from threading import Thread
from web3 import Web3


def main() -> None:

    alchemy_endpoint: Optional[str] = None
    infura_endpoint: Optional[str] = None
    eth_tx_receipt_endpoint: Optional[str] = None

    default_web3 = Web3(Web3.HTTPProvider())
    gas_limit = 22000
    parser = get_argument_parser()
    args = parser.parse_args()
    sender_private_key = args.sender_private_key
    receiver_address = default_web3.toChecksumAddress(args.receiver_address)  # pyre-ignore
    blxr_endpoint = args.blxr_endpoint
    blxr_auth_header = args.blxr_auth_header
    num_tx_groups = args.num_tx_groups
    gas_price_wei = args.gas_price * int(1e9)
    delay = args.delay

    # Check Alchemy and Infura API keys
    alchemy_api_key = args.alchemy_api_key
    if alchemy_api_key:
        alchemy_endpoint = "https://eth-mainnet.alchemyapi.io/v2/" + alchemy_api_key
        eth_tx_receipt_endpoint = alchemy_endpoint
        default_web3 = Web3(Web3.HTTPProvider(alchemy_endpoint))
        if not default_web3.isConnected():
            print(f"Alchemy endpoint {alchemy_endpoint} is not connected. Please check your API key and try again.")
            sys.exit(0)
    infura_api_key = args.infura_api_key
    if infura_api_key:
        infura_endpoint = "https://mainnet.infura.io/v3/" + infura_api_key
        eth_tx_receipt_endpoint = infura_endpoint
        default_web3 = Web3(Web3.HTTPProvider(infura_endpoint))
        if not default_web3.isConnected():
            print(f"Infura endpoint {infura_endpoint} is not connected. Please check your API key and try again.")
            sys.exit(0)

    if not (alchemy_api_key or infura_api_key):
        print("Please provide at least one API key with --alchemy-api-key or --infura-api-key.")
        sys.exit(0)

    # Check bloXroute authorization header
    blxr_quota_usage_request_json = {"method": "quota_usage"}
    result = requests.post(
        blxr_endpoint,
        json=blxr_quota_usage_request_json,
        headers={
            "Content-Type": "text/plain",
            "Authorization": blxr_auth_header
        }
    )
    result_json = result.json().encode("utf-8").decode("unicode_escape")
    result_json = json.loads(result_json)
    if "result" not in result_json or "quota_limit" not in result_json["result"]:
        print(f"Failed to check bloXroute account status with result {result_json} returned. Please provide correct "
              f"bloXroute authorization header with argument --blxr-auth-header and try again.")
        sys.exit(0)

    # Check sender's wallet balance
    sender_account = Account.from_key(sender_private_key)  # pyre-ignore
    sender_address = sender_account.address

    # pyre-ignore[29]: `web3.method.Method` is not a function.
    nonce = default_web3.eth.getTransactionCount(sender_address)
    # pyre-ignore[29]: `web3.method.Method` is not a function.
    sender_balance = default_web3.eth.getBalance(sender_address)
    sender_balance_in_eth = default_web3.fromWei(sender_balance, "ether")  # pyre-ignore

    sender_expense = num_tx_groups * gas_price_wei * gas_limit
    if sender_balance < sender_expense:
        projected_ether = default_web3.fromWei(sender_expense, "ether")  # pyre-ignore
        print(f"Sender {sender_address} does not have enough balance for {num_tx_groups} groups of transactions. "
              f"Sender's balance is {sender_balance_in_eth} ETH, "
              f"while at least {projected_ether} ETH is required")
        sys.exit(0)

    print(f"Initial check completed. Sleeping {delay} sec.")
    time.sleep(delay)

    # Generate and send num_tx_groups transactions
    group_num_to_tx_hash = {}
    for i in range(1, num_tx_groups + 1):

        provider_to_tx_hash = {}
        threads = []
        print(f"Sending tx group {i}.")

        # Alchemy transaction
        if alchemy_api_key:
            tx_alchemy = {
                "to": receiver_address,
                "value": 0,
                "gas": gas_limit,
                "gasPrice": gas_price_wei,
                "nonce": nonce,
                "chainId": 1,
                "data": "0x11"
            }
            signed_tx_alchemy = default_web3.eth.account.sign_transaction(tx_alchemy, sender_private_key)
            provider_to_tx_hash[Provider.ALCHEMY] = signed_tx_alchemy.hash.hex()
            thread_alchemy = Thread(target=send_tx_eth, args=([signed_tx_alchemy, alchemy_endpoint]))
            threads.append(thread_alchemy)

        # Infura transaction
        if infura_api_key:
            tx_infura = {
                "to": receiver_address,
                "value": 0,
                "gas": gas_limit,
                "gasPrice": gas_price_wei,
                "nonce": nonce,
                "chainId": 1,
                "data": "0x22"
            }
            signed_tx_infura = default_web3.eth.account.sign_transaction(tx_infura, sender_private_key)
            provider_to_tx_hash[Provider.INFURA] = signed_tx_infura.hash.hex()
            thread_infura = Thread(target=send_tx_eth, args=([signed_tx_infura, infura_endpoint]))
            threads.append(thread_infura)

        # bloXroute Cloud-API transaction
        tx_blxr = {
            "to": receiver_address,
            "value": 0,
            "gas": gas_limit,
            "gasPrice": gas_price_wei,
            "nonce": nonce,
            "chainId": 1,
            "data": "0x33"
        }
        signed_tx_blxr = default_web3.eth.account.sign_transaction(tx_blxr, sender_private_key)
        provider_to_tx_hash[Provider.BLOXROUTE] = signed_tx_blxr.hash.hex()
        thread_blxr = Thread(target=send_tx_blxr, args=([signed_tx_blxr, blxr_endpoint, blxr_auth_header]))
        threads.append(thread_blxr)

        # Start the threads in parallel
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

        nonce += 1
        group_num_to_tx_hash[i] = provider_to_tx_hash
        # Add a delay to all the groups except for the last group
        if i < num_tx_groups:
            print(f"Sleeping {delay} sec.")
            time.sleep(delay)

    print("Sleeping 1 min before checking transaction status.")
    time.sleep(60)

    provider_to_tx_mined = defaultdict(int)
    mined_tx_nums = set()
    sleep_left_minute = 4
    while len(mined_tx_nums) < num_tx_groups and sleep_left_minute > 0:
        for group_num in group_num_to_tx_hash:

            # Continue for confirmed transactions
            if group_num in mined_tx_nums:
                continue

            # Check transactions sent by different providers and find the confirmed one
            provider_to_tx_hash = group_num_to_tx_hash[group_num]
            for provider in provider_to_tx_hash:
                tx_hash = provider_to_tx_hash[provider]
                get_tx_receipt_request_json = {
                    "jsonrpc": "2.0", "method": "eth_getTransactionReceipt", "params": [tx_hash], "id": 1
                }
                result = requests.post(
                    eth_tx_receipt_endpoint,
                    json=get_tx_receipt_request_json,
                    headers={"Content-Type": "application/json"}
                )
                result_json = result.json()
                if "result" not in result_json:
                    print(f"Failed to check transaction status of tx {tx_hash} in group {group_num}. "
                          f"Skipping current group.")
                    break
                if result_json["result"] is not None:
                    provider_to_tx_mined[provider] += 1
                    mined_tx_nums.add(group_num)
                    break

        # When there is any pending transaction, maximum sleep time is 4 min
        if len(mined_tx_nums) < num_tx_groups:
            print(f"{num_tx_groups - len(mined_tx_nums)} transactions are pending. "
                  f"Sleeping 1 min before checking status again.")
            time.sleep(60)
            sleep_left_minute -= 1

    print("---------------------------------------------------------------------------------------------------------")
    print("Summary:")
    print(f"Sent {num_tx_groups} groups of transactions to bloXroute Cloud-API and other providers, "
          f"{len(mined_tx_nums)} of them have been confirmed: ")
    if alchemy_api_key:
        print(f"Number of Alchemy transactions mined: {provider_to_tx_mined[Provider.ALCHEMY]}")
    if infura_api_key:
        print(f"Number of Infura transactions mined: {provider_to_tx_mined[Provider.INFURA]}")
    print(f"Number of bloXrotue transactions mined: {provider_to_tx_mined[Provider.BLOXROUTE]}")


class Provider(Enum):
    ALCHEMY = "Alchemy"
    BLOXROUTE = "bloXrotue"
    INFURA = "Infura"


# Send transaction to bloXroute Cloud-API
def send_tx_blxr(tx: SignedTransaction, endpoint: str, auth_header: str) -> None:

    tx_params = {
        "transaction": tx.rawTransaction.hex()[2:],
        "synchronous": "True"
    }
    send_tx_request_json = {
        "method": "blxr_tx",
        "params": tx_params
    }

    print(f"{datetime.datetime.utcnow()} - Sending transaction to bloXroute Cloud-API. TX Hash: {tx.hash.hex()}")
    result = requests.post(
        endpoint,
        json=send_tx_request_json,
        headers={
            "Content-Type": "text/plain",
            "Authorization": auth_header
        }
    )
    result_json = result.json().encode("utf-8").decode("unicode_escape")
    result_json = json.loads(result_json)

    if "result" not in result_json or "tx_hash" not in result_json["result"]:
        print(f"{datetime.datetime.utcnow()} - Failed to send raw transaction {tx.rawTransaction.hex()} to "
              f"bloXroute Cloud-API, with result {result_json} returned. ")


# Send transaction to other service providers (Alchemy, Infura)
def send_tx_eth(tx: SignedTransaction, endpoint: str) -> None:

    params = [tx.rawTransaction.hex()]
    send_tx_request_json = {
        "jsonrpc": "2.0",
        "method": "eth_sendRawTransaction",
        "params": params,
        "id": 1
    }

    print(f"{datetime.datetime.utcnow()} - Sending transaction to {endpoint}. TX Hash: {tx.hash.hex()}")
    result = requests.post(endpoint, json=send_tx_request_json, headers={"Content-Type": "application/json"})
    result_json = result.json()

    if "result" not in result_json:
        print(f"{datetime.datetime.utcnow()} - Failed to use endpoint {endpoint} to "
              f"send raw tx {tx.rawTransaction.hex()}, with result {result_json} returned. ")


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--infura-api-key", type=str,
                        help="Infura API key / project ID. "
                             "Sample Infura endpoint: https://mainnet.infura.io/v3/<api_key>.")
    parser.add_argument("--alchemy-api-key", type=str,
                        help="Alchemy API key. "
                             "Sample Alchemy endpoint: https://eth-mainnet.alchemyapi.io/v2/<api_key>.")
    parser.add_argument("--blxr-endpoint", type=str, default="http://api.blxrbdn.com:443",
                        help="bloXroute endpoint. Use http://api.blxrbdn.com:443 for Cloud-API.")
    parser.add_argument("--blxr-auth-header", type=str, required=True,
                        help="bloXroute authorization header. Use base64 encoded value of "
                             "account_id:secret_hash for Cloud-API. For more information, see "
                             "https://bloxroute.com/docs/bloxroute-documentation/cloud-api/overview/")
    parser.add_argument("--sender-private-key", type=str, required=True,
                        help="Sender's private key, which starts with 0x.")
    parser.add_argument("--receiver-address", type=str, required=True,
                        help="Receiver's address, which starts with 0x.")
    parser.add_argument("--num-tx-groups", type=int, default=1,
                        help="Number of groups of transactions to submit.")
    parser.add_argument("--gas-price", type=int, required=True,
                        help="Transaction gas price in Gwei.")
    parser.add_argument("--delay", type=int, default=30,
                        help="Time (sec) to sleep between two consecutive groups.")
    return parser


if __name__ == "__main__":
    main()



from bxcommon.models.serializeable_enum import SerializeableEnum


class ExtensionModulesState(SerializeableEnum):
    UNAVAILABLE = "Extensions Unavailable"
    INVALID_VERSION = "Invalid Version"
    OK = "OK"



from abc import ABC
from typing import Optional, List

import blxr_rlp as rlp
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.eth.abstract_eth_message import AbstractEthMessage
from bxgateway.messages.eth.new_block_parts import NewBlockParts
from bxcommon.messages.eth.serializers.transaction import Transaction
from bxgateway.messages.eth.protocol.new_block_eth_protocol_message import NewBlockEthProtocolMessage
from bxcommon.messages.eth.serializers.block_header import BlockHeader
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_utils


class InternalEthBlockInfo(AbstractEthMessage, AbstractBlockMessage, ABC):
    """
    Internal structure that is used to transfer Ethereum new block information in BDN.
    Gateway and node may communicate new blocks using NewBlockEthProtocolMessage or NewBlockHashesEthProtocolMessage.
    This message combines information for both and used for block compression in both cases.
    The message is RLP serialized the same way as other Ethereum messages

    Message structure is: [block header] + [transactions] + [uncles] + [total difficulty] + [block number].

    Block number is 0 if the message represents NewBlockEthProtocolMessage.

    If the message represents NewBlockHashesEthProtocolMessage then block number is > 0 and total difficulty is 0.
    """

    fields = [
        ("header", BlockHeader),
        ("transactions", rlp.sedes.CountableList(Transaction)),
        ("uncles", rlp.sedes.CountableList(BlockHeader)),
        ("chain_difficulty", rlp.sedes.big_endian_int),
        ("number", rlp.sedes.big_endian_int)
    ]

    def __init__(self, msg_bytes, *args, **kwargs):
        super(InternalEthBlockInfo, self).__init__(msg_bytes, *args, **kwargs)

        self._block_header: Optional[memoryview] = None
        self._block_hash: Optional[Sha256Hash] = None
        self._timestamp: Optional[int] = None
        self._difficulty: Optional[int] = None
        self._block_number: Optional[int] = None

    def block_header(self) -> memoryview:
        if self._block_header is None:
            _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
            block_msg_bytes = self._memory_view[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

            _, block_hdr_itm_len, block_hdr_itm_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
            self._block_header = block_msg_bytes[0:block_hdr_itm_start + block_hdr_itm_len]

        block_header = self._block_header
        assert block_header is not None

        return block_header

    def block_hash(self) -> Sha256Hash:
        if self._block_hash is None:
            raw_hash = eth_common_utils.keccak_hash(self.block_header())
            self._block_hash = Sha256Hash(raw_hash)

        block_hash = self._block_hash
        assert block_hash is not None

        return block_hash

    def extra_stats_data(self) -> str:
        if self.has_block_number():
            if self.has_total_difficulty():
                return "New Block Hash msg; total difficulty calculated"
            else:
                return "New Block Hash msg; total difficulty not calculated"
        else:
            return "New Block msg"

    def prev_block_hash(self) -> Sha256Hash:
        _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
        block_msg_bytes = self._memory_view[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

        _, block_hdr_itm_len, block_hdr_itm_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
        block_hdr_bytes = block_msg_bytes[block_hdr_itm_start:block_hdr_itm_start + block_hdr_itm_len]

        _, prev_block_itm_len, prev_block_itm_start = rlp_utils.consume_length_prefix(block_hdr_bytes, 0)
        prev_block_bytes = block_hdr_bytes[prev_block_itm_start:prev_block_itm_start + prev_block_itm_len]

        return Sha256Hash(prev_block_bytes)

    def difficulty(self) -> int:
        """
        :return: seconds since epoch
        """
        if self._difficulty is None:
            _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
            block_msg_bytes = self._memory_view[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

            _, block_hdr_itm_len, block_hdr_itm_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
            block_hdr_bytes = block_msg_bytes[block_hdr_itm_start:block_hdr_itm_start + block_hdr_itm_len]

            offset = BlockHeader.FIXED_LENGTH_FIELD_OFFSET

            difficulty, _difficulty_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            self._difficulty = difficulty

        difficulty = self._difficulty
        assert difficulty is not None

        return difficulty

    def timestamp(self) -> int:
        """
        :return: seconds since epoch
        """
        if self._timestamp is None:
            _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
            block_msg_bytes = self._memory_view[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

            _, block_hdr_itm_len, block_hdr_itm_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
            block_hdr_bytes = block_msg_bytes[block_hdr_itm_start:block_hdr_itm_start + block_hdr_itm_len]

            offset = BlockHeader.FIXED_LENGTH_FIELD_OFFSET
            _difficulty, difficulty_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            offset += difficulty_length
            _number, number_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            offset += number_length
            _gas_limit, gas_limit_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            offset += gas_limit_length
            _gas_used, gas_used_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            offset += gas_used_length

            timestamp, _timestamp_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            self._timestamp = timestamp

        timestamp = self._timestamp
        assert timestamp is not None

        return timestamp

    def has_total_difficulty(self) -> bool:
        _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
        block_msg_bytes = self._memory_view[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

        _, block_header_len, block_header_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
        _, txs_len, txs_start = rlp_utils.consume_length_prefix(block_msg_bytes, block_header_start + block_header_len)
        _, uncles_len, uncles_start = rlp_utils.consume_length_prefix(block_msg_bytes, txs_start + txs_len)

        chain_difficulty, _ = rlp_utils.decode_int(block_msg_bytes, uncles_start + uncles_len)

        return chain_difficulty > 0

    def has_block_number(self) -> bool:
        return self.block_number() > 0

    def block_number(self) -> int:
        if self._block_number is None:
            _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
            block_msg_bytes = self._memory_view[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

            _, block_hdr_itm_len, block_hdr_itm_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
            block_hdr_bytes = block_msg_bytes[block_hdr_itm_start:block_hdr_itm_start + block_hdr_itm_len]

            offset = BlockHeader.FIXED_LENGTH_FIELD_OFFSET
            _difficulty, difficulty_length = rlp_utils.decode_int(block_hdr_bytes, offset)
            offset += difficulty_length
            self._block_number, _ = rlp_utils.decode_int(block_hdr_bytes, offset)

        block_number = self._block_number
        assert block_number is not None
        return block_number

    @classmethod
    def from_new_block_msg(cls, new_block_msg: NewBlockEthProtocolMessage) -> "InternalEthBlockInfo":
        """
        Creates NewBlockInternalEthMessage from raw bytes of NewBlockEthProtocolMessage
        :param new_block_msg: new block message
        :return: NewBlockInternalEthMessage message
        """
        new_block_msg_bytes = memoryview(new_block_msg.rawbytes())

        _, block_msg_itm_len, block_msg_itm_start = rlp_utils.consume_length_prefix(new_block_msg_bytes, 0)
        block_msg_bytes = new_block_msg_bytes[block_msg_itm_start:block_msg_itm_start + block_msg_itm_len]

        msg_size = 0

        # block item already include header, transactions and uncles
        _, block_itm_len, block_itm_start = rlp_utils.consume_length_prefix(block_msg_bytes, 0)
        block_itm_bytes = block_msg_bytes[block_msg_itm_start:block_msg_itm_start + block_itm_len]
        msg_size += len(block_itm_bytes)

        difficulty_bytes = block_msg_bytes[block_msg_itm_start + block_itm_len:]
        msg_size += len(difficulty_bytes)

        block_number_bytes = rlp_utils.encode_int(new_block_msg.number())
        msg_size += len(block_number_bytes)

        msg_prefix = rlp_utils.get_length_prefix_list(
            len(block_itm_bytes) + len(difficulty_bytes) + len(block_number_bytes))
        msg_size += len(msg_prefix)

        msg_bytes = bytearray(msg_size)

        written_bytes = 0

        msg_bytes[written_bytes:written_bytes + len(msg_prefix)] = msg_prefix
        written_bytes += len(msg_prefix)

        msg_bytes[written_bytes:written_bytes + len(block_itm_bytes)] = block_itm_bytes
        written_bytes += len(block_itm_bytes)

        msg_bytes[written_bytes:written_bytes + len(difficulty_bytes)] = difficulty_bytes
        written_bytes += len(difficulty_bytes)

        msg_bytes[written_bytes:written_bytes + len(block_number_bytes)] = block_number_bytes
        written_bytes += len(block_number_bytes)

        assert written_bytes == msg_size

        return cls(msg_bytes)

    @classmethod
    def from_new_block_parts(cls, new_block_details: NewBlockParts,
                             total_difficulty: Optional[int] = 0) -> "InternalEthBlockInfo":
        """
        Creates NewBlockInternalEthMessage from block header and block body bytes

        :param new_block_details: new block details
        :param total_difficulty: total difficulty of the block if known
        :return: instance of NewBlockInternalEthMessage
        """
        block_header_bytes = new_block_details.block_header_bytes
        block_body_bytes = new_block_details.block_body_bytes
        block_number = new_block_details.block_number

        if isinstance(block_body_bytes, (bytes, bytearray)):
            block_body_bytes = memoryview(block_body_bytes)

        # block body content includes transactions and uncles
        _, block_content_len, block_content_start = rlp_utils.consume_length_prefix(block_body_bytes, 0)
        block_body_conent = block_body_bytes[block_content_start:]

        msg_size = len(block_header_bytes) + len(block_body_conent)

        total_difficulty_item = rlp_utils.encode_int(0 if total_difficulty is None else total_difficulty)
        msg_size += len(total_difficulty_item)

        block_number_bytes = rlp_utils.encode_int(block_number)
        msg_size += len(block_number_bytes)

        new_block_item_prefix = rlp_utils.get_length_prefix_list(msg_size)
        msg_size += len(new_block_item_prefix)

        written_bytes = 0
        new_block_bytes = bytearray(msg_size)

        new_block_bytes[written_bytes:written_bytes + len(new_block_item_prefix)] = new_block_item_prefix
        written_bytes += len(new_block_item_prefix)

        new_block_bytes[written_bytes:written_bytes + len(block_header_bytes)] = block_header_bytes
        written_bytes += len(block_header_bytes)

        new_block_bytes[written_bytes:written_bytes + len(block_body_conent)] = block_body_conent
        written_bytes += len(block_body_conent)

        new_block_bytes[written_bytes:written_bytes + len(total_difficulty_item)] = total_difficulty_item
        written_bytes += len(total_difficulty_item)

        new_block_bytes[written_bytes:written_bytes + len(block_number_bytes)] = block_number_bytes
        written_bytes += len(block_number_bytes)

        assert msg_size == written_bytes

        return cls(new_block_bytes)

    def to_new_block_msg(self) -> NewBlockEthProtocolMessage:
        """
        Converts message to instance of NewBlockEthProtocolMessage
        :return: instance of NewBlockEthProtocolMessage
        """

        _, msg_itm_len, msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
        msg_itm_bytes = self._memory_view[msg_itm_start:]

        offset = 0
        msg_size = 0

        _, header_len, header_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        header_bytes = msg_itm_bytes[offset:header_start + header_len]
        offset = header_start + header_len
        msg_size += len(header_bytes)

        _, txs_len, txs_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        txs_bytes = msg_itm_bytes[offset:txs_start + txs_len]
        offset = txs_start + txs_len
        msg_size += len(txs_bytes)

        _, uncles_len, uncles_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        uncles_bytes = msg_itm_bytes[offset:uncles_start + uncles_len]
        offset = uncles_start + uncles_len
        msg_size += len(uncles_bytes)

        _, total_difficulty_len, total_difficulty_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        total_difficulty_bytes = msg_itm_bytes[offset:total_difficulty_start + total_difficulty_len]
        msg_size += len(total_difficulty_bytes)

        block_prefix = rlp_utils.get_length_prefix_list(
            len(header_bytes) + len(txs_bytes) + len(uncles_bytes))
        msg_size += len(block_prefix)

        msg_prefix = rlp_utils.get_length_prefix_list(msg_size)
        msg_size += len(msg_prefix)

        result_msg_bytes = bytearray(msg_size)
        written_bytes = 0

        result_msg_bytes[written_bytes:written_bytes + len(msg_prefix)] = msg_prefix
        written_bytes += len(msg_prefix)

        result_msg_bytes[written_bytes:written_bytes + len(block_prefix)] = block_prefix
        written_bytes += len(block_prefix)

        result_msg_bytes[written_bytes:written_bytes + len(header_bytes)] = header_bytes
        written_bytes += len(header_bytes)

        result_msg_bytes[written_bytes:written_bytes + len(txs_bytes)] = txs_bytes
        written_bytes += len(txs_bytes)

        result_msg_bytes[written_bytes:written_bytes + len(uncles_bytes)] = uncles_bytes
        written_bytes += len(uncles_bytes)

        result_msg_bytes[written_bytes:written_bytes + len(total_difficulty_bytes)] = total_difficulty_bytes
        written_bytes += len(total_difficulty_bytes)

        assert written_bytes == msg_size

        return NewBlockEthProtocolMessage(result_msg_bytes)

    def to_new_block_parts(self) -> NewBlockParts:
        _, msg_itm_len, msg_itm_start = rlp_utils.consume_length_prefix(self._memory_view, 0)
        msg_itm_bytes = self._memory_view[msg_itm_start:]

        offset = 0

        _, header_len, header_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        header_bytes = msg_itm_bytes[offset:header_start + header_len]
        offset = header_start + header_len

        _, txs_len, txs_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        txs_bytes = msg_itm_bytes[offset:txs_start + txs_len]
        offset = txs_start + txs_len

        _, uncles_len, uncles_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        uncles_bytes = msg_itm_bytes[offset:uncles_start + uncles_len]
        offset = uncles_start + uncles_len

        _, total_difficulty_len, total_difficulty_start = rlp_utils.consume_length_prefix(msg_itm_bytes, offset)
        offset = total_difficulty_start + total_difficulty_len

        block_number, _ = rlp_utils.decode_int(msg_itm_bytes, offset)

        block_body_prefix = rlp_utils.get_length_prefix_list(len(txs_bytes) + len(uncles_bytes))
        block_body_bytes = bytearray(len(block_body_prefix) + len(txs_bytes) + len(uncles_bytes))

        block_body_bytes[:len(block_body_prefix)] = block_body_prefix
        written_bytes = len(block_body_prefix)

        block_body_bytes[written_bytes:written_bytes + len(txs_bytes)] = txs_bytes
        written_bytes += len(txs_bytes)

        block_body_bytes[written_bytes:written_bytes + len(uncles_bytes)] = uncles_bytes
        written_bytes += len(uncles_bytes)

        return NewBlockParts(header_bytes, block_body_bytes, block_number)

    @classmethod
    def unpack(cls, buf):
        """
        Unpack buffer into command, metadata (e.g. magic number, checksum), and payload.
        """
        raise NotImplementedError()

    @classmethod
    def validate_payload(cls, buf, unpacked_args) -> None:
        """
        Validates unpacked content.
        """
        raise NotImplementedError()

    def txns(self) -> List[Transaction]:
        return self.get_field_value("transactions")



from typing import NamedTuple


class IpEndpoint(NamedTuple):
    ip_address: str
    port: int

    def __repr__(self) -> str:
        return f"{self.ip_address} {self.port}"

    def __eq__(self, other: "IpEndpoint") -> bool:
        return self.ip_address == other.ip_address and self.port == other.port

    def __hash__(self):
        return hash((self.ip_address, self.port))



from typing import List, TypeVar, Generic, Set

from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.utils.expiration_queue import ExpirationQueue
from bxutils import logging

T = TypeVar("T")
logger = logging.get_logger(__name__)


class ExpiringSet(Generic[T]):
    """
    Set with expiration time.
    """

    contents: Set[T]
    _alarm_queue: AlarmQueue
    _expiration_queue: ExpirationQueue[T]
    _expiration_time: int
    _log_removal: bool
    _name: str

    def __init__(
        self, alarm_queue: AlarmQueue, expiration_time_s: int, name: str, log_removal: bool = False
    ):
        self.contents = set()
        self._alarm_queue = alarm_queue
        self._expiration_queue = ExpirationQueue(expiration_time_s)
        self._expiration_time = expiration_time_s
        self._log_removal = log_removal
        self._name = name

    def __contains__(self, item: T) -> bool:
        return item in self.contents

    def __len__(self) -> int:
        return len(self.contents)

    def add(self, item: T) -> None:
        self.contents.add(item)
        self._expiration_queue.add(item)
        self._alarm_queue.register_approx_alarm(
            self._expiration_time * 2,
            self._expiration_time,
            self.cleanup,
            alarm_name=f"ExpiringSet[{self._name}]#cleanup"
        )

    def remove(self, item: T) -> None:
        self.contents.remove(item)

    def get_recent_items(self, count: int) -> List[T]:
        items = []
        # noinspection PyTypeChecker
        entries = reversed(self._expiration_queue.queue.keys())

        try:
            for _ in range(count):
                items.append(next(entries))
        except StopIteration as _e:
            logger.debug("Attempted to fetch {} entries, but only {} existed.", count, len(items))

        return items

    def cleanup(self) -> int:
        self._expiration_queue.remove_expired(remove_callback=self._safe_remove_item)
        return 0

    def _safe_remove_item(self, item: T):
        if self._log_removal:
            logger.debug("Removing {} from expiring set.", item)
        if item in self.contents:
            self.contents.remove(item)



from enum import IntEnum


class NotificationCodeRange(IntEnum):
    DEBUG = 2000  # <2000
    INFO = 4000  # <4000
    WARNING = 6000  # <6000
    ERROR = 8000  # <8000


class NotificationCode(IntEnum):
    QUOTA_FILL_STATUS = 4000
    ASSIGNING_SHORT_IDS = 4001
    NOT_ASSIGNING_SHORT_IDS = 4002
    ACCOUNT_EXPIRED_NOTIFICATION = 4003



import struct
from abc import ABCMeta, abstractmethod
from typing import Tuple, Optional, List, Set, Union, NamedTuple, Deque

from bxcommon import constants
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.bloxroute import compact_block_short_ids_serializer

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils.object_hash import Sha256Hash, convert
from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple

from bxgateway.utils.block_info import BlockInfo


class BlockDecompressionResult(NamedTuple):
    block_msg: Optional[AbstractBlockMessage]
    block_info: BlockInfo
    unknown_short_ids: List[int]
    unknown_tx_hashes: List[Sha256Hash]


def finalize_block_bytes(
        buf: Deque[Union[bytes, bytearray, memoryview]], size: int, short_ids: List[int]
) -> memoryview:
    serialized_short_ids = compact_block_short_ids_serializer.serialize_short_ids_into_bytes(short_ids)
    buf.append(serialized_short_ids)
    size += constants.UL_ULL_SIZE_IN_BYTES
    offset_buf = struct.pack("<Q", size)
    buf.appendleft(offset_buf)
    size += len(serialized_short_ids)

    block = bytearray(size)
    off = 0
    for blob in buf:
        next_off = off + len(blob)
        block[off:next_off] = blob
        off = next_off

    return memoryview(block)


class AbstractMessageConverter(SpecialMemoryProperties, metaclass=ABCMeta):
    """
    Message converter abstract class.

    Converts messages of specific blockchain protocol to internal messages
    """

    @abstractmethod
    def tx_to_bx_txs(
        self,
        tx_msg,
        network_num: int,
        transaction_flag: Optional[TransactionFlag] = None,
        min_tx_network_fee: int = 0,
        account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
    ) -> List[Tuple[TxMessage, Sha256Hash, Union[bytearray, memoryview]]]:
        """
        Converts blockchain transactions message to internal transaction message

        :param tx_msg: blockchain transactions message
        :param network_num: blockchain network number
        :param transaction_flag: the transaction_flag type to assign to the BDN transaction.
        :param min_tx_network_fee: minimum transaction fee. If support by the network, transactions
                                   with fees lower than this will be excluded from the result
        :param account_id: gateway's account id
        :return: array of tuples (transaction message, transaction hash, transaction bytes)
        """

        pass

    @abstractmethod
    def bx_tx_to_tx(self, bx_tx_msg):
        """
        Converts internal transaction message to blockchain transactions message

        :param bx_tx_msg: internal transaction message
        :return: blockchain transactions message
        """

        pass

    @abstractmethod
    def block_to_bx_block(
        self, block_msg, tx_service, enable_block_compression: bool, min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        """
        Convert blockchain block message to internal broadcast message with transactions replaced with short ids

        :param block_msg: blockchain new block message
        :param tx_service: Transactions service
        :param enable_block_compression
        :param min_tx_age_seconds
        :return: Internal broadcast message bytes (bytearray), tuple (txs count, previous block hash, short ids)
        """
        pass

    @abstractmethod
    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Converts internal broadcast message to blockchain new block message

        Returns None for block message if any of the transactions shorts ids or hashes are unknown

        :param bx_block_msg: internal broadcast message bytes
        :param tx_service: Transactions service
        :return: block decompression result
        """
        pass

    @abstractmethod
    def bdn_tx_to_bx_tx(
        self,
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        transaction_flag: Optional[TransactionFlag] = None,
        account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
    ) -> TxMessage:
        """
        Convert a raw transaction which arrived from an RPC request into bx transaction.
        :param raw_tx: The raw transaction bytes.
        :param network_num: the network number.
        :param transaction_flag: the quota type to assign to the BDN transaction.
        :param account_id: node's account id
        :return: bx transaction.
        """
        pass

    def encode_raw_msg(self, raw_msg: str) -> bytes:
        """
        Encode a raw message string into bytes
        :param raw_msg: the raw message to encode
        :return: binary encoded message
        :raise ValueError: if the encoding fails
        """
        return convert.hex_to_bytes(raw_msg)

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return super(AbstractMessageConverter, self).special_memory_size(ids)



from typing import Optional, Dict, List, Tuple

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway import log_messages
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.services.abstract_block_queuing_service import AbstractBlockQueuingService
from bxutils import logging

logger = logging.get_logger(__name__)


class BlockQueuingServiceManager:
    block_storage: ExpiringDict[Sha256Hash, Optional[AbstractBlockMessage]]
    blockchain_peer_to_block_queuing_service: Dict[AbstractGatewayBlockchainConnection, AbstractBlockQueuingService]
    designated_queuing_service: Optional[AbstractBlockQueuingService] = None

    def __init__(
        self,
        block_storage: ExpiringDict[Sha256Hash, Optional[AbstractBlockMessage]],
        blockchain_peer_to_block_queuing_service: Dict[AbstractGatewayBlockchainConnection, AbstractBlockQueuingService]
    ) -> None:
        self.block_storage = block_storage
        self.blockchain_peer_to_block_queuing_service = blockchain_peer_to_block_queuing_service

    def __iter__(self):
        for queuing_service in self.blockchain_peer_to_block_queuing_service.values():
            yield queuing_service

    def is_in_any_queuing_service(self, block_hash: Sha256Hash) -> bool:
        """
        :param block_hash:
        :return: if block hash is in any queuing service.
        """
        for queuing_service in self:
            if block_hash in queuing_service:
                return True
        return False

    def is_in_common_block_storage(self, block_hash: Sha256Hash) -> bool:
        """
        :param block_hash:
        :return: if block message is in common block storage for block hash
        """
        block = self.block_storage.contents.get(block_hash)
        if block is None:
            return False
        else:
            return True

    def add_block_queuing_service(
        self,
        connection: AbstractGatewayBlockchainConnection,
        block_queuing_service: AbstractBlockQueuingService
    ) -> None:
        self.blockchain_peer_to_block_queuing_service[connection] = block_queuing_service
        if self.designated_queuing_service is None:
            self.designated_queuing_service = block_queuing_service

    def remove_block_queuing_service(self, connection: AbstractGatewayBlockchainConnection) -> None:
        if connection in self.blockchain_peer_to_block_queuing_service:
            designated_queuing_service = self.designated_queuing_service
            if designated_queuing_service is not None and designated_queuing_service.connection == connection:
                self.designated_queuing_service = None
            del self.blockchain_peer_to_block_queuing_service[connection]

        for queuing_service in self.blockchain_peer_to_block_queuing_service.values():
            self.designated_queuing_service = queuing_service
            return
        return

    def get_block_queuing_service(
        self, connection: AbstractGatewayBlockchainConnection
    ) -> Optional[AbstractBlockQueuingService]:
        if connection in self.blockchain_peer_to_block_queuing_service:
            return self.blockchain_peer_to_block_queuing_service[connection]
        else:
            logger.error(log_messages.ATTEMPTED_FETCH_FOR_NONEXISTENT_QUEUING_SERVICE, connection)
            return None

    def get_designated_block_queuing_service(self) -> Optional[AbstractBlockQueuingService]:
        if self.designated_queuing_service is not None:
            return self.designated_queuing_service
        else:
            logger.error(log_messages.ATTEMPTED_FETCH_FOR_NONEXISTENT_QUEUING_SERVICE, "(designated)")
            return None

    def push(
        self,
        block_hash: Sha256Hash, block_message: Optional[AbstractBlockMessage] = None,
        waiting_for_recovery: bool = False,
        node_received_from: Optional[AbstractGatewayBlockchainConnection] = None
    ) -> None:
        for node_conn, block_queuing_service in self.blockchain_peer_to_block_queuing_service.items():
            if node_received_from and node_conn == node_received_from:
                continue
            block_queuing_service.push(block_hash, block_message, waiting_for_recovery)

    def store_block_data(self, block_hash: Sha256Hash, block_message: AbstractBlockMessage) -> None:
        if self.designated_queuing_service is not None:
            queuing_service = self.designated_queuing_service
            assert queuing_service is not None
            # Note: Stores to common block storage. Individual service used in order to call protocol specific method.
            queuing_service.store_block_data(block_hash, block_message)
        else:
            logger.error(log_messages.ATTEMPTED_FETCH_FOR_NONEXISTENT_QUEUING_SERVICE, "(designated)")

    def get_block_data(self, block_hash: Sha256Hash) -> Optional[AbstractBlockMessage]:
        if block_hash in self.block_storage:
            return self.block_storage[block_hash]
        else:
            return None

    def remove(self, block_hash: Sha256Hash) -> None:
        for queuing_service in self:
            if block_hash in queuing_service:
                queuing_service.remove(block_hash)
        if block_hash in self.block_storage:
            del self.block_storage[block_hash]

    def update_recovered_block(
        self,
        block_hash: Sha256Hash,
        block_message: AbstractBlockMessage,
        node_received_from: Optional[AbstractGatewayBlockchainConnection] = None
    ) -> None:
        self.store_block_data(block_hash, block_message)
        for node_conn, queuing_service in self.blockchain_peer_to_block_queuing_service.items():
            if node_received_from and node_conn == node_received_from:
                continue
            queuing_service.update_recovered_block(block_hash, block_message)

    def get_length_of_each_queuing_service_stats_format(self) -> str:
        queue_lengths_and_endpoints: List[Tuple[IpEndpoint, int]] = []
        for queuing_service in self:
            queue_lengths_and_endpoints.append((queuing_service.connection.endpoint, (len(queuing_service))))
        queue_lengths_str = ', '.join(
            [f"{str(ip_endpoint)}: {str(queue_length)}" for ip_endpoint, queue_length in queue_lengths_and_endpoints]
        )
        return f"[{queue_lengths_str}]"



from dataclasses import dataclass
from typing import List

from bxgateway.utils.logging.status.environment import Environment
from bxgateway.utils.logging.status.extension_modules_state import ExtensionModulesState
from bxgateway.utils.logging.status.network import Network


@dataclass
class Analysis:
    time_started: str
    startup_parameters: str
    gateway_version: str
    extensions_check: ExtensionModulesState
    environment: Environment
    network: Network
    installed_python_packages: List[str]



from typing import Optional

from bxcommon.utils.stats.stat_event_logic_flags import StatEventLogicFlags


class StatEventTypeSettings:
    def __init__(self, name: str, event_logic_flags: Optional[StatEventLogicFlags] = StatEventLogicFlags.NONE,
                 detailed_stat_event: Optional[bool] = False, priority: bool = False):
        self.name = name
        self.event_logic = event_logic_flags
        self.detailed_stat_event = detailed_stat_event
        self.priority = priority



from typing import Set, TYPE_CHECKING, cast

from bxcommon import constants
from bxcommon.utils.expiring_set import ExpiringSet
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.feed.feed import Feed
from bxcommon.feed.feed_source import FeedSource

from bxgateway import gateway_constants

from bxgateway.feed.eth.eth_block_feed_entry import EthBlockFeedEntry
from bxgateway.feed.eth.eth_raw_block import EthRawBlock
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.messages.gateway.confirmed_block_message import ConfirmedBlockMessage
from bxgateway.services.eth.eth_block_queuing_service import EthBlockQueuingService
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)

# Note: Block feed normal use case is only supported for gateways with a single blockchain connection
# i.e. Multi-node gateway users cannot assume the block has been accepted to their Ethereum instance


class EthNewBlockFeed(Feed[EthBlockFeedEntry, EthRawBlock]):
    NAME = "newBlocks"
    FIELDS = ["hash", "header", "transactions", "uncles"]
    ALL_FIELDS = FIELDS
    VALID_SOURCES = {
        FeedSource.BLOCKCHAIN_SOCKET, FeedSource.BLOCKCHAIN_RPC, FeedSource.BDN_SOCKET, FeedSource.BDN_INTERNAL
    }
    published_blocks: ExpiringSet[Sha256Hash]
    published_blocks_height: ExpiringSet[int]

    def __init__(self, node: "EthGatewayNode", network_num: int = constants.ALL_NETWORK_NUM,) -> None:
        super().__init__(self.NAME, network_num=network_num)
        self.last_block_number = 0
        self.hash_for_last_block_number = set()
        self.node = node
        self.published_blocks = ExpiringSet(
            node.alarm_queue, gateway_constants.MAX_BLOCK_CACHE_TIME_S, name="published_blocks"
        )
        self.published_blocks_height = ExpiringSet(
            node.alarm_queue, gateway_constants.MAX_BLOCK_CACHE_TIME_S, name="published_blocks_height"
        )

    def serialize(self, raw_message: EthRawBlock) -> EthBlockFeedEntry:
        block_message = raw_message.block
        assert block_message is not None
        return EthBlockFeedEntry(raw_message.block_hash, block_message)

    def publish_blocks_from_queue(self, start_block_height, end_block_height) -> Set[int]:
        missing_blocks = set()
        block_queuing_service = cast(
            EthBlockQueuingService,
            self.node.block_queuing_service_manager.get_designated_block_queuing_service()
        )
        if block_queuing_service is None:
            return missing_blocks

        for block_number in range(start_block_height, end_block_height):
            block_hash = block_queuing_service.accepted_block_hash_at_height.contents.get(block_number)
            if block_hash:
                self.publish(
                    EthRawBlock(
                        block_number,
                        block_hash,
                        FeedSource.BDN_INTERNAL,
                        self.node._get_block_message_lazy(None, block_hash)
                    )
                )
            else:
                missing_blocks.add(block_number)
        return missing_blocks

    def publish(self, raw_message: EthRawBlock) -> None:
        logger.trace(
            "attempting to publish message: {} for feed {}", raw_message, self.name
        )
        if raw_message.source not in self.VALID_SOURCES:
            return

        block_hash = raw_message.block_hash
        block_number = raw_message.block_number

        if self.node.opts.stream_to_peer_gateway is not None:
            eth_block_info = cast(InternalEthBlockInfo, raw_message.block)
            self.node.eth_ws_proxy_publisher.broadcast_confirmation_message(
                ConfirmedBlockMessage(block_hash, eth_block_info.rawbytes())
            )

        if self.subscriber_count() == 0:
            return

        if block_number < self.last_block_number - gateway_constants.MAX_BLOCK_BACKLOG_TO_PUBLISH:
            # published block is too far behind ignore
            return

        if block_hash in self.published_blocks:
            # already published ignore
            return

        if raw_message.block is None:
            block_queuing_service = cast(
                EthBlockQueuingService,
                self.node.block_queuing_service_manager.get_designated_block_queuing_service()
            )
            best_accepted_height, _ = block_queuing_service.best_accepted_block

            logger.warning(
                "{} Feed Failed to recover block for message: {},"
                "last_block_published {} last block in queueing service {}",
                self.name, raw_message, self.last_block_number, best_accepted_height
            )
            return

        self.published_blocks.add(block_hash)
        self.published_blocks_height.add(block_number)

        if self.last_block_number and block_number > self.last_block_number + 1:
            # try to publish all intermediate blocks first
            missing_blocks = self.publish_blocks_from_queue(self.last_block_number + 1, block_number - 1)
            if missing_blocks:
                logger.info("Attempting to publish to feed block: {}, missing previous blocks {} ",
                            block_number, missing_blocks
                            )

        logger.debug("{} Processing new block message: {}", self.name, raw_message)
        super(EthNewBlockFeed, self).publish(raw_message)

        if block_number in self.published_blocks_height and block_number <= self.last_block_number:
            # possible fork, try to republish all later blocks
            _missing_blocks = self.publish_blocks_from_queue(block_number + 1, self.last_block_number)

        if block_number > self.last_block_number:
            self.last_block_number = block_number


import socket
from typing import Dict, Tuple, Any

from mock import MagicMock

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.models.authenticated_peer_info import AuthenticatedPeerInfo
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.socket_connection_protocol import SocketConnectionProtocol


class MockSocketConnection(SocketConnectionProtocol):
    def __init__(
        self,
        file_no,
        node,
        default_socket_opts=None,
        send_bytes=False,
        ip_address: str = "127.0.0.1",
        port: int = 8000,
        is_ssl: bool = False,
        authenticated_peer_info: AuthenticatedPeerInfo = AuthenticatedPeerInfo(
            ConnectionType.EXTERNAL_GATEWAY, "", ""
        ),
    ):
        super(MockSocketConnection, self).__init__(
            node, IpEndpoint(ip_address, port), is_ssl=is_ssl
        )
        self.transport = MagicMock()
        if default_socket_opts is None:
            default_socket_opts = {(socket.SOL_SOCKET, socket.SO_SNDBUF): 4096}

        self.file_no = file_no
        # TODO: temporary fix for some situations where, see https://bloxroute.atlassian.net/browse/BX-1153
        self._send_bytes = send_bytes

        self.bytes_sent = []
        self.socket_opts: Dict[Tuple[int, int], Any] = default_socket_opts

        self.transport.write = self.socket_instance_send
        self.transport.get_write_buffer_size = MagicMock(return_value=0)

        self.authenticated_peer_info = authenticated_peer_info
        self.initialized = True

    def fileno(self) -> int:
        return self.file_no

    # TODO: temporary fix for some situations where, see https://bloxroute.atlassian.net/browse/BX-1153
    def send(self) -> None:
        if self._node is None or not self._send_bytes:
            return
        else:
            super(MockSocketConnection, self).send()

    def socket_instance_send(self, bytes_written):
        if isinstance(bytes_written, memoryview):
            bytes_written = bytearray(bytes_written.tobytes())
        self.bytes_sent.append(bytes_written)
        return len(bytes_written)

    def socket_instance_get_opt(self, level: int, opt_name: int):
        return self.socket_opts[(level, opt_name)]

    def socket_instance_set_opt(self, level: int, opt_name: int, value: Any):
        self.socket_opts[(level, opt_name)] = value

    def mark_for_close(self, should_retry: bool = True) -> None:
        super(MockSocketConnection, self).mark_for_close(should_retry)
        self.connection_lost(None)

    def get_write_buffer_size(self) -> int:
        return 0



from abc import abstractmethod

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils.object_hash import Sha256Hash


class AbstractBlockMessage(AbstractMessage):

    @abstractmethod
    def block_hash(self) -> Sha256Hash:
        pass

    @abstractmethod
    def prev_block_hash(self) -> Sha256Hash:
        pass

    @abstractmethod
    def timestamp(self) -> int:
        pass

    def extra_stats_data(self) -> str:
        return ""

    @abstractmethod
    def txns(self):
        # each blockchain network returns its own list format
        pass



from typing import TYPE_CHECKING

from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class OntRelayConnection(AbstractRelayConnection):

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "AbstractGatewayNode"):
        super(OntRelayConnection, self).__init__(sock, node)

    # TODO: implement msg_tx



from enum import Flag
from functools import reduce


class SerializableFlag(Flag):
    def __str__(self):
        if self.name is not None:
            return str(self.name)
        else:
            return '|'.join([str(m.name) for m in self.__class__ if m in self])

    @classmethod
    def from_string(cls, string_value: str):
        string_values = {w.strip().upper() for w in str(string_value).split("|")}
        flag_elements = [item for item in cls if item.name in string_values]
        if flag_elements:
            return reduce(lambda x, y: x | y, flag_elements)
        raise ValueError("{0} is not a valid {1}".format(string_value, cls.__name__))



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.protocol_version import PROTOCOL_VERSION
from bxcommon.messages.bloxroute.version_message import VersionMessage
from bxcommon.utils import uuid_pack
from bxcommon.utils.message_buffer_builder import PayloadElement, PayloadBlock


class HelloMessage(VersionMessage):
    """
    BloXroute relay hello message type.

    node_id: the id of the node

    """
    MESSAGE_TYPE = BloxrouteMessageType.HELLO
    HELLO_MESSAGE_BLOCK = PayloadBlock(
        VersionMessage.BASE_LENGTH,
        "HelloMessage",
        PROTOCOL_VERSION,
        PayloadElement(
            name="node_id",
            structure="%ss" % constants.NODE_ID_SIZE_IN_BYTES,
            encode=uuid_pack.to_bytes,
            decode=uuid_pack.from_bytes
        )
    )
    HELLO_MESSAGE_LENGTH = (
        VersionMessage.VERSION_MESSAGE_BLOCK.size
        + HELLO_MESSAGE_BLOCK.size
        + constants.CONTROL_FLAGS_LEN
    )

    # pyre-fixme[9]: node_id has type `str`; used as `None`.
    def __init__(self, protocol_version: Optional[int] = None, network_num: Optional[int] = None, node_id: str = None,
                 # pyre-fixme[9]: buf has type `bytearray`; used as `None`.
                 buf: bytearray = None):
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + self.HELLO_MESSAGE_LENGTH)
            buf = self.HELLO_MESSAGE_BLOCK.build(buf, node_id=node_id)

        self.buf = buf
        self._node_id = None
        self._network_num = None
        self._memoryview = memoryview(buf)
        super(HelloMessage, self).__init__(self.MESSAGE_TYPE, self.HELLO_MESSAGE_LENGTH,
                                           protocol_version, network_num, buf)

    def __unpack(self):
        contents = self.HELLO_MESSAGE_BLOCK.read(self._memoryview)
        self._node_id = contents.get("node_id")

    def node_id(self):
        if self._node_id is None:
            self.__unpack()
        return self._node_id



from enum import Enum


class BlockchainProtocol(Enum):
    BITCOIN = "bitcoin"
    BITCOINCASH = "bitcoincash"
    ETHEREUM = "ethereum"
    ONTOLOGY = "ontology"

    def __eq__(self, other) -> bool:
        return self is other or str(self) == str(other)

    def __str__(self) -> str:
        return str(self.value)

    def __hash__(self) -> int:
        return id(self)



import struct
from abc import ABC
from typing import Optional, Union

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.utils import crypto, uuid_pack
from bxcommon.utils.object_hash import Sha256Hash, ConcatHash


class AbstractBroadcastMessage(AbstractBloxrouteMessage, ABC):
    PAYLOAD_LENGTH = crypto.SHA256_HASH_LEN + constants.NETWORK_NUM_LEN + constants.NODE_ID_SIZE_IN_BYTES + \
                     constants.CONTROL_FLAGS_LEN
    SOURCE_ID_OFFSET = AbstractBloxrouteMessage.HEADER_LENGTH + crypto.SHA256_HASH_LEN + constants.NETWORK_NUM_LEN

    _message_hash: Optional[Sha256Hash] = None
    _network_num: Optional[int] = None
    _source_id: Optional[str] = None
    _message_id: Optional[ConcatHash] = None

    def __init__(
        self,
        message_hash: Optional[Sha256Hash] = None,
        network_num: Optional[int] = None,
        source_id: Optional[str] = None,
        buf: Optional[Union[bytearray, memoryview]] = None
    ):
        if buf is None:
            self.buf = bytearray(self.HEADER_LENGTH + self.PAYLOAD_LENGTH)
        else:
            self.buf = buf

        payload_length = len(self.buf) - self.HEADER_LENGTH
        super().__init__(self.MESSAGE_TYPE, payload_length, self.buf)

        if buf is None:
            assert message_hash is not None and network_num is not None and source_id is not None
            off = AbstractBloxrouteMessage.HEADER_LENGTH

            self.buf[off:off + crypto.SHA256_HASH_LEN] = message_hash.binary
            off += crypto.SHA256_HASH_LEN

            struct.pack_into("<L", self.buf, off, network_num)
            off += constants.NETWORK_NUM_LEN

            struct.pack_into("<16s", self.buf, off, uuid_pack.to_bytes(source_id))
            off += constants.NODE_ID_SIZE_IN_BYTES

    def set_message_hash(self, message_hash: Sha256Hash) -> None:
        assert self.buf is not None
        off = AbstractBloxrouteMessage.HEADER_LENGTH
        self.buf[off:off + crypto.SHA256_HASH_LEN] = message_hash.binary

    def message_hash(self) -> Sha256Hash:
        if self._message_hash is None:
            off = self.HEADER_LENGTH
            self._message_hash = Sha256Hash(self._memoryview[off:off + crypto.SHA256_HASH_LEN])

        message_hash = self._message_hash
        assert message_hash is not None
        return message_hash

    def message_id(self) -> ConcatHash:
        """
        Concatenated hash, includes network info with message hash.
        """
        if self._message_id is None:
            off = self.HEADER_LENGTH
            self._message_id = ConcatHash(
                self._memoryview[off:off + crypto.SHA256_HASH_LEN + constants.NETWORK_NUM_LEN],
                0
            )
        message_id = self._message_id
        assert message_id is not None
        return message_id

    def network_num(self) -> int:
        if self._network_num is None:
            off = self.HEADER_LENGTH + crypto.SHA256_HASH_LEN
            self._network_num, = struct.unpack_from("<L", self.buf, off)

        network_num = self._network_num
        assert network_num is not None
        return network_num

    def source_id(self) -> str:
        if self._source_id is None:
            off = self.SOURCE_ID_OFFSET
            self._source_id = uuid_pack.from_bytes(struct.unpack_from("<16s", self.buf, off)[0])
            if self._source_id is None:
                self._source_id = constants.DECODED_EMPTY_SOURCE_ID

        source_id = self._source_id
        assert source_id is not None
        return source_id

    def source_id_as_str(self) -> str:
        if self.source_id() == constants.DECODED_EMPTY_SOURCE_ID:
            return "None"
        else:
            return self.source_id()

    def set_source_id(self, source_id: str) -> None:
        self._source_id = source_id
        off = self.HEADER_LENGTH + crypto.SHA256_HASH_LEN + constants.NETWORK_NUM_LEN
        struct.pack_into("<16s", self.buf, off, uuid_pack.to_bytes(source_id))

    def has_source_id(self) -> bool:
        return self.source_id() != constants.DECODED_EMPTY_SOURCE_ID



import struct
from collections import namedtuple

from bxcommon import constants

BlockOffsets = namedtuple("BlockOffsets", ["short_id_offset", "block_begin_offset"])


def get_serialized_short_ids_bytes_len(short_ids) -> int:
    """
    Calculates length of list of short ids serialized into bytes

    :param short_ids: list of short ids
    :return: length of serialized bytes
    """

    return constants.UL_INT_SIZE_IN_BYTES + len(short_ids) * constants.UL_INT_SIZE_IN_BYTES


def serialize_short_ids_into_bytes(short_ids) -> bytearray:
    """
    Serializes short ids into bytes

    :param short_ids: list of short ids
    :return: bytearray with serialized bytes
    """

    buffer = bytearray(get_serialized_short_ids_bytes_len(short_ids))

    struct.pack_into("<L", buffer, 0, len(short_ids))
    offset = constants.UL_INT_SIZE_IN_BYTES

    for short_id in short_ids:
        struct.pack_into("<L", buffer, offset, short_id)
        offset += constants.UL_INT_SIZE_IN_BYTES

    return buffer


def deserialize_short_ids_from_buffer(buffer, offset):
    """
    Deserializes list of short ids from buffer

    :param buffer: buffer containing serialized short ids bytes
    :param offset: offset in a buffer where serialized short ids begin
    :return: list of short ids
    """

    short_ids_len, = struct.unpack_from("<L", buffer, offset)
    length = constants.UL_INT_SIZE_IN_BYTES

    short_ids = []

    for _ in range(short_ids_len):
        short_id, = struct.unpack_from("<L", buffer, offset + length)
        length += constants.UL_INT_SIZE_IN_BYTES
        short_ids.append(short_id)

    return short_ids, length


def get_bx_block_offsets(bx_block):
    short_id_offset, = struct.unpack_from("<Q", bx_block)
    return BlockOffsets(short_id_offset, constants.UL_ULL_SIZE_IN_BYTES)



from collections import deque
from typing import Deque, Union
from typing import Set, Optional

from bxcommon.utils import memory_utils
from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple


class InputBuffer(SpecialMemoryProperties):
    def __init__(self) -> None:
        self.input_list: Deque[Union[memoryview, bytearray, bytes]] = deque()
        self.length = 0

    def endswith(self, suffix: Union[memoryview, bytearray, bytes]) -> bool:
        if not self.input_list:
            return False

        # pyre-fixme[25]: Assertion will always fail.
        if not isinstance(suffix, (memoryview, bytearray, bytes)):
            raise ValueError(f"Suffix must be memoryview, bytearray or bytes, not {type(suffix)}.")

        end_slice = self.input_list[-1]
        if len(end_slice) < len(suffix):
            end_slice = self.get_slice(self.length - len(suffix), self.length)
        if not isinstance(end_slice, bytearray):
            end_slice = bytearray(end_slice)
            self.input_list[-1] = end_slice

        return end_slice.endswith(suffix)  # pyre-ignore

    def add_bytes(self, piece: Union[bytearray, bytes, memoryview]) -> None:
        """
        Adds a data bytes to the end of the input buffer.
        """
        # pyre-fixme[25]: Assertion will always fail.
        if not isinstance(piece, (bytearray, memoryview, bytes)):
            raise ValueError("Piece must be a bytearray, bytes or memoryview.")
        self.input_list.append(piece)
        self.length += len(piece)

    def remove_bytes(self, num_bytes: int) -> Union[bytearray, bytes, memoryview]:
        """
        Removes the first num_bytes bytes in the input buffer and returns them.
        """
        if num_bytes is None or num_bytes < 0:
            raise ValueError("Invalid num_bytes {}".format(num_bytes))

        self._shrink_if_needed(num_bytes, True)

        assert self.input_list or num_bytes == 0, f"Input buffer is empty and attempting to remove {num_bytes} bytes!"

        if self.input_list and num_bytes > 0:
            to_return = self.input_list[0][:num_bytes]
            self.input_list[0] = self.input_list[0][num_bytes:]
            self.length -= num_bytes
        else:
            to_return = bytearray(0)

        return to_return

    def peek_message(self, bytes_to_peek):
        """
        Returns at LEAST the first bytes_to_peek bytes in the input buffer.
        The assumption is that these bytes are all part of the same message.
        Thus, we combine pieces if we cannot just return the first message.
        """
        if bytes_to_peek > self.length:
            bytes_to_peek = self.length

        if bytes_to_peek == 0:
            return bytearray(0)
        self._shrink_if_needed(bytes_to_peek)

        return self.input_list[0]

    def get_slice(self, start, end):
        """
        Gets a slice of the inputbuffer from start to end.
        We assume that this slice is a piece of a single bitcoin message
        for performance reasons (with respect to the number of copies).
        Additionally, the start value of the slice must exist.
        """
        if start is None or end is None or self.length < start:
            raise ValueError("Start ({}) and end ({}) must exist and start must be less or equal to length ({})."
                             .format(start, end, self.length))

        # Combine all of the pieces in this slice into the first item on the list.
        # Since we will need to do so anyway when handing the message.
        self._shrink_if_needed(end)
        return self.input_list[0][start:end]

    def __len__(self) -> int:
        return self.length

    def __getitem__(self, item) -> Union[memoryview, bytearray, bytes]:
        if not isinstance(item, slice):
            raise ValueError("Input buffer does not support nonslice indexing")

        if item.step is not None and item.step != 1:
            raise ValueError("Input buffer does not support non 1 slice step values")

        start = item.start
        if start is None:
            start = 0

        stop = item.stop
        if stop is None:
            stop = self.length

        return self.get_slice(start, stop)

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return memory_utils.get_special_size(self.input_list, ids=ids)

    def _shrink_if_needed(self, end: int, ensure_byte_array: bool = False) -> None:
        if self.input_list:
            head_message = self.input_list[0]
            if end > len(head_message) or ensure_byte_array:
                if isinstance(head_message, bytearray):
                    head = self.input_list.popleft()
                else:
                    head = bytearray(self.input_list.popleft())
                while end > len(head) and self.input_list:
                    head.extend(self.input_list.popleft())  # pyre-ignore
                self.input_list.appendleft(head)



import struct

from bxcommon.exceptions import ParseError
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxgateway.utils.eth import frame_utils
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_constants
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher


class FramedInputBuffer(object):
    """
    Input buffer for Ethereum framed messages
    """

    def __init__(self, rlpx_cipher):

        if not isinstance(rlpx_cipher, RLPxCipher):
            raise TypeError("Cipher is expected of type RLPxCipher but was {0}"
                            .format(type(rlpx_cipher)))

        self._rlpx_cipher = rlpx_cipher
        self._payload_buffer = InputBuffer()

        self._receiving_frame = False
        self._chunked_frames_in_progress = False
        self._chunked_frames_total_body_size = False
        self._chunked_frames_body_size_received = 0

        self._current_msg_type = None

        self._current_frame_size = None
        self._current_frame_body_size = None
        self._current_frame_enc_body_size = None
        self._current_frame_protocol_id = None
        self._current_frame_sequence_id = None

        self._full_message_received = False

    def peek_message(self, input_buffer):
        """
        Peeks message from input frame
        :param input_buffer: input buffer
        :return: tuple (flag if full message is received, message type)
        """

        if not isinstance(input_buffer, InputBuffer):
            raise ValueError("Expected type InputBuffer")

        if self._full_message_received:
            raise ValueError("Get full message before trying to peek another one")

        if not self._receiving_frame and input_buffer.length >= eth_common_constants.FRAME_HDR_TOTAL_LEN:
            enc_header_bytes = input_buffer.remove_bytes(eth_common_constants.FRAME_HDR_TOTAL_LEN)
            header_bytes = self._rlpx_cipher.decrypt_frame_header(enc_header_bytes)
            body_size, protocol_id, sequence_id, total_payload_len = frame_utils.parse_frame_header(header_bytes)

            self._current_frame_body_size = body_size
            self._current_frame_protocol_id = protocol_id
            self._current_frame_size = frame_utils.get_full_frame_size(body_size)
            self._current_frame_enc_body_size = self._current_frame_size - eth_common_constants.FRAME_HDR_TOTAL_LEN
            self._current_frame_sequence_id = sequence_id

            if sequence_id == 0 and total_payload_len is not None:
                self._chunked_frames_in_progress = True
                self._chunked_frames_total_body_size = total_payload_len

            self._receiving_frame = True

        if self._receiving_frame and input_buffer.length >= self._current_frame_enc_body_size:
            frame_enc_body_bytes = input_buffer.remove_bytes(self._current_frame_enc_body_size)

            body = self._rlpx_cipher.decrypt_frame_body(frame_enc_body_bytes, self._current_frame_body_size)

            msg_type_is_expected = not self._chunked_frames_in_progress or self._current_frame_sequence_id == 0
            payload, msg_type = frame_utils.parse_frame_body(body, msg_type_is_expected)

            if msg_type_is_expected:
                self._current_msg_type = msg_type

            self._payload_buffer.add_bytes(bytearray(rlp_utils.str_to_bytes(payload)))

            if not self._chunked_frames_in_progress:
                self._full_message_received = True
            else:
                self._chunked_frames_body_size_received += len(body)

                if self._chunked_frames_body_size_received > self._chunked_frames_total_body_size:
                    raise ParseError("Expected total body length for frame message is {0} but received {1}"
                                     .format(self._chunked_frames_total_body_size,
                                             self._chunked_frames_body_size_received))

                if self._chunked_frames_body_size_received == self._chunked_frames_total_body_size:
                    self._full_message_received = True

            self._receiving_frame = False

        return self._full_message_received, self._current_msg_type

    def get_full_message(self):
        """
        Returns full message from input buffer
        :return: full message
        """

        assert self._full_message_received

        message_length = self._payload_buffer.length
        message = self._payload_buffer.remove_bytes(message_length)
        msg_type = self._current_msg_type

        self._full_message_received = False

        self._receiving_frame = False
        self._chunked_frames_in_progress = False
        self._chunked_frames_total_body_size = False
        self._chunked_frames_body_size_received = 0

        self._current_msg_type = None
        self._current_frame_size = None
        self._current_frame_body_size = None
        self._current_frame_enc_body_size = None
        self._current_frame_protocol_id = None
        self._current_frame_sequence_id = None

        return message, msg_type

    def peek_eip8_handshake_message_len(self, input_buffer: InputBuffer) -> int:
        handshake_message_len = 0
        if input_buffer.length > eth_common_constants.EIP8_AUTH_PREFIX_LEN:
            cipher_text = input_buffer.peek_message(eth_common_constants.EIP8_AUTH_PREFIX_LEN)
            handshake_message_len = struct.unpack(">H", cipher_text[:eth_common_constants.EIP8_AUTH_PREFIX_LEN])[0] + \
                                    eth_common_constants.EIP8_AUTH_PREFIX_LEN
        return handshake_message_len



from typing import Optional

from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

logger_filters = logging.get_logger(LogRecordType.TransactionFiltering, __name__)


def reformat_tx_value(value: str) -> float:
    return float(int(value, 0))


def reformat_address(address: str) -> str:
    # eth allows 0x in the `to` when a new contract was created. In this case we pad the 0
    # the issue is in th pycond that search strings as substring and not exact
    if address == "0x":
        return "0x0000000000000000000000000000000000000000"
    return address.lower()


def reformat_gas_price(value: Optional[str]) -> int:
    # value may be None in certain scenarios
    # (e.g. type 2 transaction will have None gas_price, others None max_priority_fee_per_gas)
    if value is None:
        return -1
    return int(value, 0)


def reformat_input_to_method_id(input_field: str) -> str:
    if len(input_field) >= 10:
        return input_field[2:10]
    return ""



from dataclasses import dataclass
from typing import Optional

from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.messages.eth.serializers.block_header import BlockHeader
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_utils


@dataclass
class NewBlockParts:
    block_header_bytes: memoryview
    block_body_bytes: memoryview
    block_number: int

    def get_block_hash(self) -> Optional[Sha256Hash]:
        if self.block_header_bytes is None:
            return None

        raw_hash = eth_common_utils.keccak_hash(self.block_header_bytes)
        return Sha256Hash(raw_hash)

    def get_previous_block_hash(self) -> Optional[Sha256Hash]:
        if self.block_header_bytes is None:
            return None

        _, header_items_len, header_items_start = rlp_utils.consume_length_prefix(self.block_header_bytes, 0)
        header_items_bytes = self.block_header_bytes[header_items_start:]

        _, prev_block_itm_len, prev_block_itm_start = rlp_utils.consume_length_prefix(header_items_bytes, 0)
        prev_block_bytes = header_items_bytes[prev_block_itm_start:prev_block_itm_start + prev_block_itm_len]

        return Sha256Hash(prev_block_bytes)

    def get_block_difficulty(self) -> Optional[int]:
        if self.block_header_bytes is None:
            return None

        _, header_items_len, header_items_start = rlp_utils.consume_length_prefix(self.block_header_bytes, 0)
        header_items_bytes = self.block_header_bytes[header_items_start:]

        block_difficulty, _ = rlp_utils.decode_int(header_items_bytes, BlockHeader.FIXED_LENGTH_FIELD_OFFSET)

        return block_difficulty



from typing import Union, cast, Optional

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.bloxroute import compact_block_short_ids_serializer
from bxcommon.messages.eth.validation.eth_block_validator import EthBlockValidator
from bxcommon.utils.blockchain_utils.eth import rlp_utils
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.messages.eth.protocol.get_block_bodies_eth_protocol_message import (
    GetBlockBodiesEthProtocolMessage,
)
from bxgateway.messages.eth.protocol.get_block_headers_eth_protocol_message import (
    GetBlockHeadersEthProtocolMessage,
)
from bxgateway.services.block_processing_service import BlockProcessingService
from bxgateway.services.eth.eth_block_queuing_service import EthBlockQueuingService
from bxutils import logging

logger = logging.get_logger(__name__)


class EthBlockProcessingService(BlockProcessingService):
    _node: "bxgateway.connections.eth.eth_gateway_node.EthGatewayNode"

    def __init__(self, node):
        super(EthBlockProcessingService, self).__init__(node)

        self._block_validator = EthBlockValidator()

    def try_process_get_block_headers_request(
        self, msg: GetBlockHeadersEthProtocolMessage, block_queuing_service: Optional[EthBlockQueuingService]
    ) -> bool:
        if block_queuing_service is None:
            return False

        block_hash = msg.get_block_hash()

        if block_hash is not None:
            logger.trace(
                "Checking for headers by hash ({}) in local block cache...",
                block_hash,
            )
            (
                success, requested_block_hashes
            ) = block_queuing_service.get_block_hashes_starting_from_hash(
                block_hash,
                msg.get_amount(),
                msg.get_skip(),
                bool(msg.get_reverse()),
            )
        else:
            block_number = msg.get_block_number()
            if block_number:
                logger.trace(
                    "Checking for headers by block number ({}) "
                    "in local block cache",
                    block_number,
                )
                (
                    success, requested_block_hashes
                ) = block_queuing_service.get_block_hashes_starting_from_height(
                    block_number,
                    msg.get_amount(),
                    msg.get_skip(),
                    bool(msg.get_reverse()),
                )
            else:
                logger.debug(
                    "Unexpectedly, request for headers did not contain "
                    "block hash or block number. Skipping."
                )
                return False

        if success:
            return block_queuing_service.try_send_headers_to_node(
                requested_block_hashes
            )
        else:
            logger.trace(
                "Could not find requested block hashes. "
                "Forwarding to remote blockchain connection."
            )
            return False

    def try_process_get_block_bodies_request(
        self, msg: GetBlockBodiesEthProtocolMessage, block_queuing_service: Optional[EthBlockQueuingService]
    ) -> bool:
        if block_queuing_service is None:
            return False

        block_hashes = msg.get_block_hashes()
        logger.trace("Checking for bodies in local block cache...")

        return block_queuing_service.try_send_bodies_to_node(
            block_hashes
        )

    def _get_compressed_block_header_bytes(self, compressed_block_bytes: Union[bytearray, memoryview]) -> Union[
        bytearray, memoryview]:
        block_msg_bytes = compressed_block_bytes if isinstance(compressed_block_bytes, memoryview) else memoryview(compressed_block_bytes)

        block_offsets = compact_block_short_ids_serializer.get_bx_block_offsets(block_msg_bytes)

        block_bytes = block_msg_bytes[block_offsets.block_begin_offset: block_offsets.short_id_offset]

        _, block_itm_len, block_itm_start = rlp_utils.consume_length_prefix(block_bytes, 0)
        block_itm_bytes = block_bytes[block_itm_start:]

        _, block_hdr_len, block_hdr_start = rlp_utils.consume_length_prefix(block_itm_bytes, 0)
        return block_itm_bytes[0:block_hdr_start + block_hdr_len]

    def _get_block_header_bytes_from_block_message(self, block_message: AbstractBlockMessage) -> Union[
        bytearray, memoryview]:
        eth_block_message = cast(InternalEthBlockInfo, block_message)
        return eth_block_message.block_header()



from bxgateway.messages.ont.ont_normal_message_converter import OntNormalMessageConverter


def create_ont_message_converter(magic, opts):
    if opts.use_extensions or opts.import_extensions:
        from bxgateway.messages.ont.ont_extension_message_converter import OntExtensionMessageConverter

    if opts.use_extensions:
        return OntExtensionMessageConverter(magic)
    else:
        return OntNormalMessageConverter(magic)



from typing import List, Any, Union, Dict, Optional

import humps

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.json_rpc_request import JsonRpcRequest
from bxcommon.rpc.rpc_errors import RpcMethodNotFound
from bxcommon.rpc.rpc_request_type import RpcRequestType
from bxutils.encoding.json_encoder import Case


class BxJsonRpcRequest(JsonRpcRequest):
    method: RpcRequestType
    account_cache_key: Optional[str] = None

    def __init__(
        self,
        request_id: Optional[str],
        method: RpcRequestType,
        params: Union[List[Any], Dict[Any, Any], None],
    ) -> None:
        super().__init__(request_id, method.name.lower(), params)
        self.method = method

    def __str__(self) -> str:
        return f"BxJsonRpcRequest<{self.to_json()}>"

    def to_json(self, case: Case = Case.SNAKE) -> Dict[str, Any]:
        params = self.params
        if params is not None and case == Case.CAMEL:
            params = humps.camelize(params)

        return {
            rpc_constants.JSON_RPC_VERSION_FIELD: self.json_rpc_version,
            rpc_constants.JSON_RPC_REQUEST_ID: self.id,
            rpc_constants.JSON_RPC_METHOD: self.method.name.lower(),
            rpc_constants.JSON_RPC_PARAMS: params,
        }

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "BxJsonRpcRequest":
        request_id = payload.get(rpc_constants.JSON_RPC_REQUEST_ID, None)
        if rpc_constants.JSON_RPC_METHOD not in payload:
            raise RpcMethodNotFound(request_id, "RPC request does not contain a method.")

        str_method = payload[rpc_constants.JSON_RPC_METHOD]
        try:
            method = RpcRequestType[str_method.upper()]
        except KeyError:
            possible_values = [rpc_type.lower() for rpc_type in RpcRequestType.__members__.keys()]
            raise RpcMethodNotFound(
                request_id,
                f"RPC method: {str_method} is not recognized (possible_values: {possible_values}).",
            )
        request_params = payload.get(rpc_constants.JSON_RPC_PARAMS, None)
        return cls(request_id, method, request_params)



from typing import List

import blxr_rlp as rlp
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxcommon.messages.eth.serializers.transaction import Transaction


class TransactionsEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.TRANSACTIONS

    fields = [("transactions", rlp.sedes.CountableList(Transaction))]

    def __repr__(self):
        # Calling get_transactions here causes transactions message to be deserialized in Python code and impacts
        # performance. Print only length of the message instead.
        return f"TransactionsEthProtocolMessage<message_len: {len(self.rawbytes())}>"

    def get_transactions(self) -> List[Transaction]:
        return self.get_field_value("transactions")



from cryptography import utils
from cryptography.x509.extensions import UnrecognizedExtension

from bxutils.ssl.extensions.extensions_object_ids import ExtensionsObjectIds


class AccountIdExtension(UnrecognizedExtension):

    def __init__(self, account_id: str) -> None:
        super(AccountIdExtension, self).__init__(ExtensionsObjectIds.ACCOUNT_ID, account_id.encode("utf-8"))
        self._account_id = account_id

    account_id = utils.read_only_property("_account_id")

    def __repr__(self):
        return f"{self.__class__.__name__} <account_id: {self._account_id}>"

    def __eq__(self, other) -> bool:
        if not isinstance(other, AccountIdExtension):
            return NotImplemented

        return self._account_id == other._account_id

    def __ne__(self, other) -> bool:
        return not self == other

    def __hash__(self):
        return hash(self._account_id)



import dataclasses
import functools

from dataclasses import dataclass
from collections import defaultdict
from typing import Dict, Optional


@dataclass
class PeerStatsMessage:
    bytes: int = 0
    count: int = 0


@dataclass
class PeerStats:
    address: str = ""
    peer_id: Optional[str] = None
    messages_received: Dict[str, PeerStatsMessage] = dataclasses.field(
        default_factory=functools.partial(defaultdict, PeerStatsMessage)
    )
    messages_sent: PeerStatsMessage = dataclasses.field(default_factory=PeerStatsMessage)
    peer_total_received: int = 0
    peer_total_sent: int = 0
    ping_max: float = 0
    ping_incoming_max: float = 0
    ping_outgoing_max: float = 0



import enum
from typing import Iterable

from bxcommon.models.serializable_flag import SerializableFlag


class NodeType(SerializableFlag):
    INTERNAL_GATEWAY = enum.auto()
    EXTERNAL_GATEWAY = enum.auto()
    # deprecated, use GATEWAY_TYPE for abstract gateway,
    # GATEWAY for v1.0 gateways
    GATEWAY = INTERNAL_GATEWAY | EXTERNAL_GATEWAY
    GATEWAY_TYPE = GATEWAY
    RELAY_TRANSACTION = enum.auto()
    RELAY_BLOCK = enum.auto()
    RELAY = RELAY_TRANSACTION | RELAY_BLOCK
    RELAY_TYPE = RELAY  # use RELAY_TYPE for abstract relay, use RELAY for `non split` relay
    RELAY_PROXY = enum.auto()
    RELAY_GROUP = RELAY | RELAY_PROXY
    API = enum.auto()
    API_SOCKET = enum.auto()
    BLOXROUTE_CLOUD_API = enum.auto()
    JOBS = enum.auto()

    def get_subtypes(self) -> Iterable["NodeType"]:
        for subtype in self.__class__:
            if subtype in self:
                yield subtype



import struct
from typing import Optional

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.ont_messages_util import ipaddr_to_ontbytearray


class AddrOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.ADDRESS

    def __init__(self, magic: Optional[int] = None, addrs: Optional[list] = None, buf: Optional[bytearray] = None):
        if addrs is None:
            addrs = []

        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + len(addrs) * ont_constants.ONT_NODE_ADDR_LEN + 8)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<Q", buf, off, len(addrs))
            off += ont_constants.ONT_LONG_LONG_LEN

            for addr in addrs:
                timestamp, service_type, host_ip, sync_port, cons_port, pid = addr
                struct.pack_into("<QQ", buf, off, timestamp, service_type)
                off += ont_constants.ONT_ADDR_TIME_AND_SERV_LEN
                # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for
                #  2nd param but got `Optional[bytearray]`.
                buf[off:off + ont_constants.ONT_IP_ADDR_SIZE] = ipaddr_to_ontbytearray(host_ip)
                off += ont_constants.ONT_IP_ADDR_SIZE
                struct.pack_into("<HHQ", buf, off, sync_port, cons_port, pid)
                off += ont_constants.ONT_ADDR_PORTS_AND_ID_LEN

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None
